import datetime
import logging
from unittest import mock

import pytest
import smpplib
import time_machine

from canaille.app import mask_phone
from canaille.core.auth import AuthenticationSession
from canaille.core.models import OTP_VALIDITY
from canaille.core.models import SEND_NEW_OTP_DELAY
from canaille.core.sms import send_one_time_password_sms


@pytest.fixture
def configuration(configuration):
    configuration["CANAILLE"]["AUTHENTICATION_FACTORS"] = ["sms"]
    return configuration


def test_no_auth_session_not_logged_in(testclient, user):
    """Non-logged users should not be able to access the sms auth form if an authentication session has not been properly started."""
    res = testclient.get("/auth/sms", status=302)
    assert res.location == "/login"
    assert res.flashes == [
        ("warning", "Cannot remember the login you attempted to sign in with.")
    ]


def test_no_auth_session_logged_in(testclient, logged_user):
    """Logged users should not be able to access the sms auth form if an authentication session has not been properly started."""
    res = testclient.get("/auth/sms", status=302)
    assert res.location == "/"


def test_not_sms_step(testclient, user):
    """Users reaching the sms passcode form while this is not the right auth step in their flow should be redirected there."""
    with testclient.session_transaction() as session:
        session["auth"] = AuthenticationSession(
            user_name="user",
            remaining=["password", "sms"],
        ).serialize()

    res = testclient.get("/auth/sms", status=302)
    assert res.location == "/auth/password"


def test_signin_with_sms_otp(smpp_client, testclient, backend, user, caplog):
    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    assert (
        "canaille",
        logging.SECURITY,
        "Sent one-time passcode for user at 555-000-000",
    ) in caplog.record_tuples

    backend.reload(user)

    smpp_client.send_message.assert_called()
    assert (
        user.one_time_password
        in smpp_client.send_message.call_args.kwargs["short_message"].decode()
    )

    res.form["otp"] = user.one_time_password
    res = res.form.submit(status=302, name="action", value="confirm")

    assert (
        "success",
        "Connection successful. Welcome Johnny",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Successful sms code authentication for user",
    ) in caplog.record_tuples

    res = res.follow()
    assert (
        "canaille",
        logging.SECURITY,
        "Successful authentication for user",
    ) in caplog.record_tuples


def test_signin_wrong_sms_otp(testclient, user, caplog):
    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    assert (
        "canaille",
        logging.SECURITY,
        "Sent one-time passcode for user at 555-000-000",
    ) in caplog.record_tuples

    res.form["otp"] = "123456"
    res = res.form.submit(status=200, name="action", value="confirm")

    assert (
        "error",
        "The passcode you entered is invalid. Please try again",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Failed sms code authentication for user",
    ) in caplog.record_tuples


def test_expired_sms_otp(testclient, user, caplog):
    """Good expired passcode should be refused."""
    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False) as traveller:
        with testclient.session_transaction() as session:
            assert not session.get("sessions")

        res = testclient.get("/login", status=200)

        res.form["login"] = "user"
        res = res.form.submit(status=302)
        res = res.follow(status=200)

        otp = user.generate_sms_or_mail_otp()
        send_one_time_password_sms(user.preferred_email, otp)
        res.form["otp"] = user.one_time_password

        traveller.shift(datetime.timedelta(seconds=OTP_VALIDITY))
        res = res.form.submit(status=200, name="action", value="confirm")

        assert (
            "error",
            "The passcode you entered is invalid. Please try again",
        ) in res.flashes
        assert (
            "canaille",
            logging.SECURITY,
            "Failed sms code authentication for user",
        ) in caplog.record_tuples


def test_send_new_mail_otp_too_early(smpp_client, testclient, backend, user, caplog):
    """Check that clicking on the resend button too early should not send a new mail."""
    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get("/login", status=200)
        res.form["login"] = "user"
        res = res.form.submit(status=302)
        res = res.follow(status=200)
        assert smpp_client.send_message.call_count == 1

        res = res.form.submit(status=302, name="action", value="resend")

        assert (
            "danger",
            f"Too many attempts. Please try again in {SEND_NEW_OTP_DELAY} seconds.",
        ) in res.flashes
        assert smpp_client.send_message.call_count == 1
        res = res.follow(status=200)


def test_send_new_sms_otp_on_time(smpp_client, testclient, backend, user, caplog):
    """Check that clicking on the resend button after an adequate delay send a new code."""
    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get("/login", status=200)
        res.form["login"] = "user"
        res = res.form.submit(status=302)
        res = res.follow(status=200)
        assert smpp_client.send_message.call_count == 1

    with time_machine.travel("2020-01-01 01:00:11+00:00", tick=False):
        res = res.form.submit(status=302, name="action", value="resend")

        assert smpp_client.send_message.call_count == 2

        backend.reload(user)
        sms_content = smpp_client.send_message.call_args.kwargs[
            "short_message"
        ].decode()
        assert user.one_time_password in sms_content
        assert (
            "success",
            "The new verification code have been sent.",
        ) in res.flashes
        assert (
            "canaille",
            logging.SECURITY,
            "Sent one-time passcode for user at 555-000-000",
        ) in caplog.record_tuples

        assert res.location == "/auth/sms"


def test_send_new_sms_invalid_user(smpp_client, testclient, backend, user, caplog):
    """Invalid users should be shown the re-send button, but it should do nothing."""
    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get("/login", status=200)
        res.form["login"] = "invalid"
        res = res.form.submit(status=302)
        res = res.follow(status=200)
        assert not smpp_client.send_message.called

    with time_machine.travel("2020-01-01 01:00:11+00:00", tick=False):
        res = res.form.submit(status=302, name="action", value="resend")

        assert not smpp_client.send_message.called
        assert (
            "success",
            "The new verification code have been sent.",
        ) in res.flashes
        assert (
            "canaille",
            logging.SECURITY,
            "Sent one-time passcode for user at 555-000-000",
        ) not in caplog.record_tuples


def test_send_new_sms_error(smpp_client, testclient, backend, user, caplog):
    """Handle SMS sending errors."""
    smpp_client.connect = mock.Mock(
        side_effect=smpplib.exceptions.ConnectionError("Host unreachable")
    )

    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get("/login", status=200)
        res.form["login"] = "user"
        res = res.form.submit(status=302)
        res = res.follow(status=200)

    with time_machine.travel("2020-01-01 01:01:00+00:00", tick=False):
        res = res.form.submit(status=302, name="action", value="resend")
        assert (
            "canaille",
            logging.WARNING,
            "Could not send sms: Host unreachable",
        ) in caplog.record_tuples


def test_mask_phone_number():
    phone = "+33123456789"
    assert mask_phone(phone) == "+33#####89"



from urllib.parse import parse_qs
from urllib.parse import urlsplit

from joserfc import jwt

from canaille.app import models


def test_oauth_hybrid(testclient, backend, user, client):
    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code token",
            client_id=client.client_id,
            scope="openid profile",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
    ).follow()
    assert "text/html" == res.content_type, res.json

    res.form["login"] = user.user_name
    res = res.form.submit(status=302).follow()

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302).follow()

    assert "text/html" == res.content_type, res.json

    res = res.form.submit(name="answer", value="accept", status=302)

    assert res.location.startswith(client.redirect_uris[0])
    params = parse_qs(urlsplit(res.location).fragment)

    code = params["code"][0]
    authcode = backend.get(models.AuthorizationCode, code=code)
    assert authcode is not None

    access_token = params["access_token"][0]
    token = backend.get(models.Token, access_token=access_token)
    assert token is not None

    res = testclient.get(
        "/oauth/userinfo",
        headers={"Authorization": f"Bearer {access_token}"},
        status=200,
    )
    assert res.json["name"] == "John (johnny) Doe"


def test_oidc_hybrid(
    testclient, backend, logged_user, client, server_jwk, trusted_client
):
    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code id_token token",
            client_id=client.client_id,
            scope="openid profile",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
    )
    assert "text/html" == res.content_type, res.json

    res = res.form.submit(name="answer", value="accept", status=302)

    assert res.location.startswith(client.redirect_uris[0])
    params = parse_qs(urlsplit(res.location).fragment)

    code = params["code"][0]
    authcode = backend.get(models.AuthorizationCode, code=code)
    assert authcode is not None

    access_token = params["access_token"][0]
    token = backend.get(models.Token, access_token=access_token)
    assert token is not None

    id_token = params["id_token"][0]
    claims = jwt.decode(id_token, server_jwk)
    assert logged_user.user_name == claims.claims["sub"]
    assert logged_user.formatted_name == claims.claims["name"]
    assert [client.client_id, trusted_client.client_id] == claims.claims["aud"]

    res = testclient.get(
        "/oauth/userinfo",
        headers={"Authorization": f"Bearer {access_token}"},
        status=200,
    )
    assert res.json["name"] == "John (johnny) Doe"



"""Client 'jwk' has been renamed 'jwks'.

Revision ID: 1738929510
Revises: 1736443538
Create Date: 2025-02-07 12:58:30.855043
"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "1738929510"
down_revision: str | None = "1736443538"
branch_labels: str | Sequence[str] | None = ()
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("client", sa.Column("jwks", sa.String(), nullable=True))
    op.drop_column("client", "jwk")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("client", sa.Column("jwk", sa.VARCHAR(), nullable=True))
    op.drop_column("client", "jwks")
    # ### end Alembic commands ###



import datetime
import json
import unittest.mock
from unittest import mock

import pytest

from canaille.app import models
from canaille.backends import Backend
from canaille.backends import ModelEncoder
from canaille.commands import cli


def test_serialize(user):
    """Test ModelSerializer with basic types."""
    assert json.dumps({"foo": "bar"}, cls=ModelEncoder) == '{"foo": "bar"}'

    assert (
        json.dumps({"foo": datetime.datetime(1970, 1, 1)}, cls=ModelEncoder)
        == '{"foo": "1970-01-01T00:00:00"}'
    )

    with pytest.raises(TypeError):
        json.dumps({"foo": object()}, cls=ModelEncoder)


def test_get_list_models(cli_runner, backend, user):
    """Nominal case test for model get command."""
    res = cli_runner.invoke(cli, ["get"], catch_exceptions=False)
    assert res.exit_code == 2, res.stderr
    models = ("user", "group")
    for model in models:
        assert model in res.stderr


def test_get(cli_runner, backend, user):
    """Nominal case test for model get command."""
    res = cli_runner.invoke(cli, ["get", "user"], catch_exceptions=False)
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == [
        {
            "created": mock.ANY,
            "display_name": "Johnny",
            "emails": [
                "john@doe.test",
            ],
            "family_name": "Doe",
            "formatted_address": "1234, some street, 6789 some city, some state",
            "formatted_name": "John (johnny) Doe",
            "given_name": "John",
            "id": user.id,
            "last_modified": mock.ANY,
            "password": mock.ANY,
            "phone_numbers": [
                "555-000-000",
            ],
            "preferred_language": "en",
            "profile_url": "https://john.test",
            "user_name": "user",
            "locality": "some city",
            "postal_code": "6789",
            "region": "some state",
            "street": "1234, some street",
            "hotp_counter": 1,
            "last_otp_login": mock.ANY,
            "secret_token": mock.ANY,
        },
    ]


def test_get_model_filter(cli_runner, backend, user, admin, foo_group):
    """Test model get filter."""
    res = cli_runner.invoke(
        cli, ["get", "user", "--groups", foo_group.id], catch_exceptions=False
    )
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == [
        {
            "created": mock.ANY,
            "display_name": "Johnny",
            "emails": [
                "john@doe.test",
            ],
            "family_name": "Doe",
            "formatted_address": "1234, some street, 6789 some city, some state",
            "formatted_name": "John (johnny) Doe",
            "given_name": "John",
            "id": user.id,
            "last_modified": mock.ANY,
            "password": mock.ANY,
            "phone_numbers": [
                "555-000-000",
            ],
            "preferred_language": "en",
            "profile_url": "https://john.test",
            "user_name": "user",
            "groups": [foo_group.id],
            "locality": "some city",
            "postal_code": "6789",
            "region": "some state",
            "street": "1234, some street",
            "hotp_counter": 1,
            "last_otp_login": mock.ANY,
            "secret_token": mock.ANY,
        },
    ]


def test_get_datetime_filter(cli_runner, backend, user):
    """Test model get filter."""
    res = cli_runner.invoke(
        cli,
        ["get", "user", "--created", user.created.isoformat()],
    )
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == [
        {
            "created": mock.ANY,
            "display_name": "Johnny",
            "emails": [
                "john@doe.test",
            ],
            "family_name": "Doe",
            "formatted_address": "1234, some street, 6789 some city, some state",
            "formatted_name": "John (johnny) Doe",
            "given_name": "John",
            "id": user.id,
            "last_modified": mock.ANY,
            "password": mock.ANY,
            "phone_numbers": [
                "555-000-000",
            ],
            "preferred_language": "en",
            "profile_url": "https://john.test",
            "user_name": "user",
            "locality": "some city",
            "postal_code": "6789",
            "region": "some state",
            "street": "1234, some street",
            "hotp_counter": 1,
            "last_otp_login": mock.ANY,
            "secret_token": mock.ANY,
        },
    ]


def test_get_ignore_errors(cli_runner, backend):
    """Test that errors are ignored with --ignore-errors."""

    def mock_query(*args, **kwargs):
        raise Exception("Simulated query error")

    with unittest.mock.patch.object(Backend.instance, "query", side_effect=mock_query):
        res = cli_runner.invoke(cli, ["get", "user"])
        assert res.exit_code != 0
        assert "Simulated query error" in str(res.output)

    with unittest.mock.patch.object(Backend.instance, "query", side_effect=mock_query):
        res = cli_runner.invoke(cli, ["get", "user", "--ignore-errors"])
        assert res.exit_code == 0
        assert res.stdout == ""


def test_get_with_boolean_false(cli_runner, backend):
    """Test that boolean false values work correctly in get command filters."""
    client_false = models.Client(
        client_id="test-client-false",
        client_name="Test Client False",
        require_auth_time=False,
    )
    client_true = models.Client(
        client_id="test-client-true",
        client_name="Test Client True",
        require_auth_time=True,
    )
    backend.save(client_false)
    backend.save(client_true)

    res = cli_runner.invoke(
        cli, ["get", "client", "--require-auth-time", "true"], catch_exceptions=False
    )
    assert res.exit_code == 0, res.stdout
    clients = json.loads(res.stdout)
    assert len(clients) == 1
    assert clients[0]["require_auth_time"] is True
    assert clients[0]["client_id"] == "test-client-true"

    res = cli_runner.invoke(
        cli, ["get", "client", "--require-auth-time", "false"], catch_exceptions=False
    )
    assert res.exit_code == 0, res.stdout
    clients = json.loads(res.stdout)
    assert len(clients) == 1
    assert clients[0]["require_auth_time"] is False
    assert clients[0]["client_id"] == "test-client-false"

    backend.delete(client_false)
    backend.delete(client_true)



import datetime

from flask import url_for
from scim2_models import Meta

from canaille.app import models
from canaille.backends import Backend
from canaille.scim.models import EnterpriseUser
from canaille.scim.models import Group
from canaille.scim.models import User


def user_from_canaille_to_scim(user, user_class, enterprise_user_class):
    scim_user_class = user_class if user_class != User else User[EnterpriseUser]
    scim_user = scim_user_class(
        meta=Meta(
            resource_type="User",
            created=user.created,
            last_modified=user.last_modified,
            location=url_for("scim.query_user", user=user, _external=True),
        ),
        user_name=user.user_name,
        preferred_language=user.preferred_language,
        name=user_class.Name(
            formatted=user.formatted_name,
            family_name=user.family_name,
            given_name=user.given_name,
        )
        if (user.formatted_name or user.family_name or user.given_name)
        else None,
        display_name=user.display_name,
        password=user.get_password_hash(),
        title=user.title,
        profile_url=user.profile_url,
        emails=[
            user_class.Emails(
                value=email,
            )
            for email in user.emails or []
        ]
        or None,
        phone_numbers=[
            user_class.PhoneNumbers(
                value=phone_number,
            )
            for phone_number in user.phone_numbers or []
        ]
        or None,
        addresses=[
            user_class.Addresses(
                formatted=user.formatted_address,
                street_address=user.street,
                postal_code=user.postal_code,
                locality=user.locality,
                region=user.region,
            )
        ]
        if (
            user.formatted_address
            or user.street
            or user.postal_code
            or user.locality
            or user.region
        )
        else None,
        photos=[
            user_class.Photos(
                value=url_for(
                    "core.account.photo", user=user, field="photo", _external=True
                ),
                primary=True,
                type=user_class.Photos.Type.photo,
            )
        ]
        if user.photo
        else None,
        groups=[
            user_class.Groups(
                value=group.id,
                display=group.display_name,
                ref=url_for("scim.query_group", group=group, _external=True),
            )
            for group in user.groups or []
        ]
        or None,
        active=user.lock_date is None,
    )
    if enterprise_user_class:
        scim_user[enterprise_user_class] = enterprise_user_class(
            employee_number=user.employee_number,
            organization=user.organization,
            department=user.department,
        )
    return scim_user


def user_from_canaille_to_scim_server(user):
    scim_user = user_from_canaille_to_scim(user, User, EnterpriseUser)
    scim_user.id = user.id
    return scim_user


def user_from_scim_to_canaille(scim_user: User, user):
    user.user_name = scim_user.user_name
    if scim_user.password != user.get_password_hash():
        user.password = scim_user.password
    user.preferred_language = scim_user.preferred_language
    user.formatted_name = scim_user.name.formatted if scim_user.name else None
    user.family_name = scim_user.name.family_name if scim_user.name else None
    user.given_name = scim_user.name.given_name if scim_user.name else None
    user.display_name = scim_user.display_name
    user.title = scim_user.title
    user.profile_url = scim_user.profile_url
    user.emails = [email.value for email in scim_user.emails or []] or None
    user.phone_numbers = [
        phone.value for phone in scim_user.phone_numbers or []
    ] or None

    if scim_user.addresses:
        address = scim_user.addresses[0]
        user.formatted_address = address.formatted
        user.street = address.street_address
        user.postal_code = address.postal_code
        user.locality = address.locality
        user.region = address.region
    else:
        user.formatted_address = None
        user.street = None
        user.postal_code = None
        user.locality = None
        user.region = None
    # TODO: delete the photo
    # if scim_user.photos and scim_user.photos[0].value:
    #    user.photo = scim_user.photos[0].value
    user.employee_number = (
        scim_user[EnterpriseUser].employee_number if scim_user[EnterpriseUser] else None
    )
    user.organization = (
        scim_user[EnterpriseUser].organization if scim_user[EnterpriseUser] else None
    )
    user.department = (
        scim_user[EnterpriseUser].department if scim_user[EnterpriseUser] else None
    )
    user.groups = [
        Backend.instance.get(models.Group, group.value)
        for group in scim_user.groups or []
        if group.value
    ]
    if not scim_user.active:
        user.lock_date = datetime.datetime.now(datetime.timezone.utc)
    else:
        user.lock_date = None
    return user


def group_from_canaille_to_scim(group, group_class):
    return group_class(
        meta=Meta(
            resource_type="Group",
            created=group.created,
            last_modified=group.last_modified,
            location=url_for("scim.query_group", group=group, _external=True),
        ),
        display_name=group.display_name,
    )


def group_from_canaille_to_scim_server(group):
    scim_group = group_from_canaille_to_scim(group, Group)
    scim_group.id = group.id

    scim_group.members = [
        Group.Members(
            value=user.identifier,
            display=user.display_name,
            ref=url_for("scim.query_user", user=user, _external=True),
        )
        for user in group.members or []
    ] or None
    return scim_group


def group_from_scim_to_canaille(scim_group: Group, group):
    group.display_name = scim_group.display_name

    members = []
    for member in scim_group.members or []:
        if user := Backend.instance.get(models.User, member.value):
            members.append(user)

    group.members = members

    return group



"""Tests for OIDC client branding during authentication flow."""

import uuid

from canaille.app import models


def test_standard_login_displays_generic_branding(testclient, user):
    """Verify standard login page displays generic site branding without client-specific elements."""
    with testclient.session_transaction() as sess:
        sess.clear()

    res = testclient.get("/login")

    assert "Privacy policy" not in res.text
    assert "Terms of service" not in res.text
    assert "Sign in at Canaille" in res.text
    assert "Sign in to " not in res.text


def test_client_branding_elements_display(testclient, user, client, backend):
    """Verify all client branding elements (logo, name, legal links) appear in authentication interface."""
    with testclient.session_transaction() as sess:
        sess.clear()

    client.client_name = "My Cool App"
    client.logo_uri = "https://example.com/logo.png"
    client.client_uri = "https://example.com"
    client.policy_uri = "https://example.com/privacy"
    client.tos_uri = "https://example.com/terms"
    backend.save(client)

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
        status=302,
    )

    res = res.follow()

    assert "My Cool App" in res.text
    assert "Sign in to My Cool App" in res.text
    assert client.logo_uri in res.text
    assert "Privacy policy" in res.text
    assert "Terms of service" in res.text


def test_reauthentication_prompt_displays_contextual_messaging(
    testclient, logged_user, client, backend
):
    """Verify prompt=login parameter triggers re-authentication with appropriate user messaging."""
    client.client_name = "Test Client"
    backend.save(client)

    consent = models.Consent(
        consent_id=str(uuid.uuid4()),
        client=client,
        subject=logged_user,
        scope=["openid", "profile"],
    )
    backend.save(consent)

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile",
            nonce="somenonce",
            prompt="login",
            redirect_uri="https://client.test/redirect1",
        ),
        status=302,
    )

    res = res.follow()

    assert "Re-authenticate for Test Client" in res.text
    assert "You are already logged in" in res.text
    assert "is requesting identity confirmation" in res.text

    backend.delete(consent)


def test_password_step_maintains_client_branding(testclient, user, client, backend):
    """Verify client branding persists throughout password authentication step."""
    with testclient.session_transaction() as sess:
        sess.clear()

    client.client_name = "My Test App"
    client.logo_uri = "https://example.com/logo.png"
    backend.save(client)

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
        status=302,
    )

    res = res.follow()
    res.form["login"] = user.user_name
    res = res.form.submit(status=302).follow()

    assert res.template == "core/auth/password.html"
    assert "My Test App" in res.text
    assert "Sign in to My Test App" in res.text
    assert client.logo_uri in res.text



from unittest import mock

from joserfc import jwt

from canaille.oidc.configuration import UserInfoMappingSettings
from canaille.oidc.oauth import generate_user_claims


def test_generate_user_claims(user, foo_group):
    assert generate_user_claims(user) == {
        "sub": "user",
        "name": "John (johnny) Doe",
        "given_name": "John",
        "family_name": "Doe",
        "preferred_username": "Johnny",
        "locale": "en",
        "website": "https://john.test",
        "address": {
            "formatted": "1234, some street, 6789 some city, some state",
            "locality": "some city",
            "postal_code": "6789",
            "region": "some state",
            "street_address": "1234, some street",
        },
        "email": "john@doe.test",
        "phone_number": "555-000-000",
        "groups": ["foo"],
        "updated_at": user.last_modified.timestamp(),
    }


def test_get_userinfo(testclient, token, user, foo_group, backend):
    token.scope = ["openid"]
    backend.save(token)
    res = testclient.get(
        "/oauth/userinfo",
        headers={"Authorization": f"Bearer {token.access_token}"},
        status=200,
    )
    assert res.json == {
        "sub": "user",
    }


def test_get_userinfo_jwt(testclient, token, user, client, backend, server_jwk):
    client.userinfo_signed_response_alg = "RS256"
    backend.save(client)

    res = testclient.get(
        "/oauth/userinfo",
        headers={
            "Authorization": f"Bearer {token.access_token}",
            "Content-Type": "application/jwt",
        },
        status=200,
    )
    assert res.headers["Content-Type"] == "application/jwt"
    token = jwt.decode(res.text, server_jwk)
    assert token.claims == {
        "sub": "user",
        "name": "John (johnny) Doe",
        "given_name": "John",
        "family_name": "Doe",
        "preferred_username": "Johnny",
        "website": "https://john.test",
        "locale": "en",
        "updated_at": mock.ANY,
        "iss": "http://canaille.test",
        "aud": client.client_id,
    }


def test_post_userinfo(testclient, token, user, foo_group, backend):
    token.scope = ["openid"]
    backend.save(token)
    res = testclient.post(
        "/oauth/userinfo",
        headers={"Authorization": f"Bearer {token.access_token}"},
        status=200,
    )
    assert res.json == {
        "sub": "user",
    }


def test_userinfo_scopes(testclient, token, user, foo_group, backend):
    token.scope = ["openid"]
    backend.save(token)
    res = testclient.get(
        "/oauth/userinfo",
        headers={"Authorization": f"Bearer {token.access_token}"},
        status=200,
    )
    assert res.json == {
        "sub": "user",
    }

    token.scope = ["openid", "profile"]
    backend.save(token)
    res = testclient.get(
        "/oauth/userinfo",
        headers={"Authorization": f"Bearer {token.access_token}"},
    )
    assert res.json == {
        "sub": "user",
        "given_name": "John",
        "family_name": "Doe",
        "name": "John (johnny) Doe",
        "preferred_username": "Johnny",
        "locale": "en",
        "website": "https://john.test",
        "updated_at": user.last_modified.timestamp(),
    }

    token.scope = ["openid", "profile", "email"]
    backend.save(token)
    res = testclient.get(
        "/oauth/userinfo",
        headers={"Authorization": f"Bearer {token.access_token}"},
    )
    assert res.json == {
        "sub": "user",
        "given_name": "John",
        "family_name": "Doe",
        "name": "John (johnny) Doe",
        "preferred_username": "Johnny",
        "locale": "en",
        "website": "https://john.test",
        "email": "john@doe.test",
        "updated_at": user.last_modified.timestamp(),
    }

    token.scope = ["openid", "profile", "address"]
    backend.save(token)
    res = testclient.get(
        "/oauth/userinfo",
        headers={"Authorization": f"Bearer {token.access_token}"},
    )
    assert res.json == {
        "sub": "user",
        "family_name": "Doe",
        "given_name": "John",
        "name": "John (johnny) Doe",
        "preferred_username": "Johnny",
        "locale": "en",
        "website": "https://john.test",
        "address": {
            "formatted": "1234, some street, 6789 some city, some state",
            "locality": "some city",
            "postal_code": "6789",
            "region": "some state",
            "street_address": "1234, some street",
        },
        "updated_at": user.last_modified.timestamp(),
    }

    token.scope = ["openid", "profile", "phone"]
    backend.save(token)
    res = testclient.get(
        "/oauth/userinfo",
        headers={"Authorization": f"Bearer {token.access_token}"},
    )
    assert res.json == {
        "sub": "user",
        "given_name": "John",
        "family_name": "Doe",
        "name": "John (johnny) Doe",
        "preferred_username": "Johnny",
        "locale": "en",
        "website": "https://john.test",
        "phone_number": "555-000-000",
        "updated_at": user.last_modified.timestamp(),
    }

    token.scope = ["openid", "profile", "groups"]
    backend.save(token)
    res = testclient.get(
        "/oauth/userinfo",
        headers={"Authorization": f"Bearer {token.access_token}"},
    )
    assert res.json == {
        "sub": "user",
        "given_name": "John",
        "family_name": "Doe",
        "name": "John (johnny) Doe",
        "preferred_username": "Johnny",
        "locale": "en",
        "website": "https://john.test",
        "groups": ["foo"],
        "updated_at": user.last_modified.timestamp(),
    }


def test_generate_user_standard_claims_with_default_config(testclient, backend, user):
    user.preferred_language = "fr"

    default_jwt_mapping = UserInfoMappingSettings().model_dump()
    data = generate_user_claims(user, default_jwt_mapping)

    assert data == {
        "address": {
            "formatted": "1234, some street, 6789 some city, some state",
            "locality": "some city",
            "postal_code": "6789",
            "region": "some state",
            "street_address": "1234, some street",
        },
        "sub": "user",
        "name": "John (johnny) Doe",
        "family_name": "Doe",
        "given_name": "John",
        "preferred_username": "Johnny",
        "email": "john@doe.test",
        "locale": "fr",
        "phone_number": "555-000-000",
        "website": "https://john.test",
        "updated_at": mock.ANY,
    }


def test_custom_config_format_claim_is_well_formated(testclient, backend, user):
    userinfo_mapping_config = UserInfoMappingSettings().model_dump()
    userinfo_mapping_config["EMAIL"] = "{{ user.user_name }}@mydomain.test"

    data = generate_user_claims(user, userinfo_mapping_config)

    assert data["email"] == "user@mydomain.test"


def test_claim_is_omitted_if_empty(testclient, backend, user):
    # According to https://openid.net/specs/openid-connect-core-1_0.html#UserInfoResponse
    # it's better to not insert a null or empty string value
    user.emails = []
    backend.save(user)

    default_userinfo_mapping = UserInfoMappingSettings().model_dump()
    data = generate_user_claims(user, default_userinfo_mapping)

    assert "email" not in data



from flask import g
from flask import session as flask_session
from itsdangerous.url_safe import URLSafeSerializer

from canaille.app import models
from canaille.app.session import LOGIN_HISTORY
from canaille.app.session import LOGIN_HISTORY_COOKIE
from canaille.app.session import _get_login_history_from_cookie
from canaille.app.session import add_to_login_history
from canaille.app.session import get_login_history
from canaille.app.session import login_user
from canaille.app.session import remove_from_login_history


def test_add_to_login_history_empty_identifier(testclient):
    """Test that empty identifiers are ignored."""
    with testclient.session_transaction() as sess:
        sess.clear()

    add_to_login_history("")
    add_to_login_history(None)

    with testclient.session_transaction() as sess:
        assert LOGIN_HISTORY not in sess


def test_add_to_login_history_single_user(testclient):
    """Test adding a single user to past users list."""
    with testclient.session_transaction() as sess:
        sess.clear()

    with testclient.app.test_request_context():
        flask_session.clear()
        add_to_login_history("testuser")
        assert flask_session[LOGIN_HISTORY] == ["testuser"]


def test_add_to_login_history_multiple_users(testclient):
    """Test adding multiple users maintains MRU order."""
    with testclient.app.test_request_context():
        flask_session.clear()

        add_to_login_history("user1")
        add_to_login_history("user2")
        add_to_login_history("user3")

        assert g.login_history_needs_update == ["user3", "user2", "user1"]


def test_add_to_login_history_duplicate_moves_to_front(testclient):
    """Test that re-adding an existing user moves it to front."""
    with testclient.app.test_request_context():
        add_to_login_history("user1")
        add_to_login_history("user2")
        add_to_login_history("user3")
        add_to_login_history("user1")

        assert g.login_history_needs_update == ["user1", "user3", "user2"]


def test_add_to_login_history_max_limit(testclient):
    """Test that past users list is capped at MAX_LOGIN_HISTORY."""
    with testclient.app.test_request_context():
        for i in range(8):
            add_to_login_history(f"user{i}")

        assert len(g.login_history_needs_update) == 6
        assert g.login_history_needs_update == [
            "user7",
            "user6",
            "user5",
            "user4",
            "user3",
            "user2",
        ]


def test_get_login_history_empty(testclient):
    """Test get_login_history with empty session."""
    with testclient.session_transaction() as sess:
        sess.clear()

    login_history = get_login_history()
    assert login_history == []


def test_get_login_history_with_valid_user(testclient, user):
    """Test get_login_history retrieves valid user objects."""
    with testclient.session_transaction() as sess:
        sess.clear()

    add_to_login_history(user.user_name)
    login_history = get_login_history()

    assert len(login_history) == 1
    assert login_history[0].user_name == user.user_name


def test_get_login_history_filters_deleted_users(testclient, user):
    """Test get_login_history filters out deleted/invalid users."""
    with testclient.app.test_request_context():
        add_to_login_history(user.user_name)
        add_to_login_history("nonexistent_user")

        login_history = get_login_history()

        assert len(login_history) == 1
        assert login_history[0].user_name == user.user_name


def test_login_page_without_login_history(testclient):
    """Test login page renders normally without past users."""
    with testclient.session_transaction() as sess:
        sess.clear()

    res = testclient.get("/login")
    assert res.status_code == 200
    assert "Choose an account" not in res.text
    assert "Use another account" not in res.text


def test_login_page_with_login_history(testclient, user):
    """Test login page displays past users when available."""
    with testclient.session_transaction() as sess:
        sess.clear()
        sess[LOGIN_HISTORY] = [user.user_name]

    res = testclient.get("/login")
    assert res.status_code == 200
    assert "Choose an account" in res.text
    assert "Use another account" in res.text
    assert user.name in res.text


def test_login_with_username_valid_user(testclient, user):
    """Test direct login with valid username redirects to auth."""
    res = testclient.get(f"/login/{user.user_name}")
    assert res.status_code == 302
    assert "/auth/password" in res.location


def test_login_with_username_invalid_user(testclient):
    """Test direct login with invalid username continues to auth (no user enumeration)."""
    res = testclient.get("/login/nonexistent")
    assert res.status_code == 302
    assert "/auth/password" in res.location


def test_login_with_username_invalid_user_auth_fails(testclient):
    """Test that authentication fails for invalid user without revealing it doesn't exist.

    Note: Currently, the auth templates display the username, which could allow
    user enumeration. However, this follows the existing Canaille UX pattern
    and the main protection is that we don't immediately reject invalid users.
    """
    res = testclient.get("/login/nonexistent")
    res = res.follow()
    assert res.status_code == 200
    assert res.template == "core/auth/password.html"

    res.form["password"] = "any password"
    res = res.form.submit()

    assert res.status_code == 200
    assert "Login failed" in res.text or "Authentication failed" in res.text
    assert "User not found" not in res.text


def test_login_success_adds_to_login_history(testclient, user, backend):
    """Test successful login adds user to past users list."""
    with testclient.session_transaction() as sess:
        sess.clear()

    res = testclient.get("/login")
    res.form["login"] = user.user_name
    res = res.form.submit()

    res = res.follow()
    res.form["password"] = "correct horse battery staple"
    res = res.form.submit()

    with testclient.session_transaction() as sess:
        assert LOGIN_HISTORY in sess
        assert user.user_name in sess[LOGIN_HISTORY]


def test_logout_preserves_login_history(testclient, user, backend):
    """Test logout preserves login history in cookie."""
    with testclient.session_transaction() as sess:
        sess.clear()

    res = testclient.get("/login")
    res.form["login"] = user.user_name
    res = res.form.submit()
    res = res.follow()
    res.form["password"] = "correct horse battery staple"
    res = res.form.submit()

    testclient.get("/logout")

    res = testclient.get("/login")
    assert "Choose an account" in res.text
    assert user.name in res.text


def test_login_history_card_links(testclient, user):
    """Test that past user cards have correct links."""
    with testclient.session_transaction() as sess:
        sess.clear()
        sess[LOGIN_HISTORY] = [user.user_name]

    res = testclient.get("/login")
    assert res.status_code == 200

    expected_link = f"/login/{user.user_name}"
    assert f'href="{expected_link}"' in res.text


def test_get_login_history_lookup_consistency(testclient, user):
    """Test get_login_history uses consistent lookup with get_user_from_login."""
    with testclient.session_transaction() as sess:
        sess.clear()

    add_to_login_history(user.user_name)

    login_history = get_login_history()
    assert len(login_history) == 1
    assert login_history[0] == user


def test_auth_template_displays_user_name_for_known_users(testclient, user):
    """Test that auth templates display user name only for users in LOGIN_HISTORY."""
    with testclient.session_transaction() as sess:
        sess[LOGIN_HISTORY] = [user.user_name]

    res = testclient.get(f"/login/{user.user_name}")
    res = res.follow()
    assert res.status_code == 200
    assert res.template == "core/auth/password.html"

    assert user.name in res.text
    assert user.user_name in res.text


def test_auth_template_displays_identifier_for_unknown_users(testclient, user):
    """Test that auth templates display only identifier for unknown users."""
    with testclient.session_transaction() as sess:
        sess.clear()

    res = testclient.get(f"/login/{user.user_name}")
    res = res.follow()
    assert res.status_code == 200
    assert res.template == "core/auth/password.html"

    assert user.name not in res.text
    assert user.user_name in res.text


def test_impersonated_users_not_in_login_history(testclient, logged_admin, user):
    """Test that impersonated users do not appear in login_history list."""
    with testclient.session_transaction() as sess:
        sess.pop(LOGIN_HISTORY, None)

    testclient.get(f"/impersonate/{user.user_name}", status=302)

    with testclient.session_transaction() as sess:
        login_history = sess.get(LOGIN_HISTORY, [])
        assert user.user_name not in login_history


def test_login_user_remember_parameter(testclient, user):
    """Test that login_user respects the remember parameter."""
    with testclient.app.test_request_context():
        flask_session.clear()
        login_user(user)
        assert flask_session.get(LOGIN_HISTORY, []) == [user.user_name]

        flask_session.clear()
        login_user(user, remember=False)
        assert LOGIN_HISTORY not in flask_session


def test_forget(testclient, user):
    """Test the remove past user route."""
    res = testclient.get(f"/forget/{user.user_name}")
    assert res.status_code == 302
    assert res.location.endswith("/login")


def test_forget_nonexistent_user(testclient):
    """Test removing a non-existent user returns redirect (security)."""
    with testclient.session_transaction() as sess:
        sess[LOGIN_HISTORY] = ["existing_user"]

    res = testclient.get("/forget/nonexistent")
    assert res.status_code == 302
    assert res.location.endswith("/login")

    with testclient.session_transaction() as sess:
        login_history = sess.get(LOGIN_HISTORY, [])
        assert login_history == ["existing_user"]


def test_remove_from_login_history_function(testclient):
    """Test the remove_from_login_history function directly."""
    with testclient.app.test_request_context():
        add_to_login_history("user1")
        add_to_login_history("user2")
        add_to_login_history("user3")

        remove_from_login_history("user2")
        assert g.login_history_needs_update == ["user3", "user1"]

        g.login_history_needs_update = ["user1"]
        remove_from_login_history("user1")
        assert g.login_history_needs_update == []

        g.login_history_needs_update = ["user1"]
        remove_from_login_history("nonexistent")
        assert g.login_history_needs_update == ["user1"]

        g.login_history_needs_update = ["user1"]
        remove_from_login_history("")
        assert g.login_history_needs_update == ["user1"]


def test_remember_me_checkbox_default_true(testclient):
    """Test that remember checkbox is checked by default."""
    res = testclient.get("/login")
    assert res.status_code == 200

    assert 'name="remember"' in res.text
    assert "checked" in res.text


def test_remember_me_false_no_login_history(testclient, user, backend):
    """Test that remember=False doesn't add user to login history."""
    with testclient.session_transaction() as sess:
        sess.clear()

    res = testclient.get("/login")
    res.form["login"] = user.user_name
    res.form["remember"].checked = False
    res = res.form.submit()

    res = res.follow()
    res.form["password"] = "correct horse battery staple"
    res = res.form.submit()

    testclient.get("/logout")

    res = testclient.get("/login")
    assert "Choose an account" not in res.text


def test_remember_me_true_adds_to_login_history(testclient, user, backend):
    """Test that remember=True adds user to login history."""
    with testclient.session_transaction() as sess:
        sess.clear()

    res = testclient.get("/login")
    res.form["login"] = user.user_name
    res = res.form.submit()

    res = res.follow()
    res.form["password"] = "correct horse battery staple"
    res = res.form.submit()

    testclient.get("/logout")

    res = testclient.get("/login")
    assert "Choose an account" in res.text
    assert user.name in res.text


def test_remember_me_false_preserves_existing_history(testclient, user, backend):
    """Test that remember=False doesn't remove existing login history."""
    res = testclient.get("/login")
    res.form["login"] = user.user_name
    res = res.form.submit()
    res = res.follow()
    res.form["password"] = "correct horse battery staple"
    res = res.form.submit()
    testclient.get("/logout")

    res = testclient.get("/login")
    assert "Choose an account" in res.text

    res = testclient.get("/login")
    res.form["login"] = user.user_name
    res.form["remember"].checked = False
    res = res.form.submit()
    res = res.follow()
    res.form["password"] = "correct horse battery staple"
    res = res.form.submit()
    testclient.get("/logout")

    res = testclient.get("/login")
    assert "Choose an account" in res.text
    assert user.name in res.text


def test_user_without_password_redirects_to_firstlogin_via_direct_url(
    testclient, backend
):
    """Test that direct login URL for user without password redirects to firstlogin."""
    temp_user = models.User(
        formatted_name="Temp User",
        family_name="Temp",
        user_name="tempuser",
        emails=["temp@example.com"],
    )
    backend.save(temp_user)

    res = testclient.get("/login/tempuser")

    assert res.status_code == 302
    assert res.location == "/firstlogin/tempuser"

    backend.delete(temp_user)


def test_get_login_history_from_cookie_no_cookie(testclient):
    """Test _get_login_history_from_cookie with no cookie present."""
    with testclient.app.test_request_context():
        result = _get_login_history_from_cookie()
        assert result == []


def test_get_login_history_from_cookie_invalid_data_type(testclient):
    """Test _get_login_history_from_cookie with non-list data in cookie."""
    with testclient.app.test_request_context():
        serializer = URLSafeSerializer(testclient.app.secret_key)
        invalid_data = serializer.dumps("not_a_list")

        with testclient.app.test_request_context(
            environ_base={"HTTP_COOKIE": f"{LOGIN_HISTORY_COOKIE}={invalid_data}"}
        ):
            result = _get_login_history_from_cookie()
            assert result == []


def test_get_login_history_from_cookie_corrupted_cookie(testclient):
    """Test _get_login_history_from_cookie with corrupted cookie."""
    with testclient.app.test_request_context(
        environ_base={"HTTP_COOKIE": f"{LOGIN_HISTORY_COOKIE}=corrupted_data"}
    ):
        result = _get_login_history_from_cookie()
        assert result == []



from canaille.app.configuration import BaseModel


class UserInfoMappingSettings(BaseModel):
    """Mapping between the user model and the JWT fields.

    Fields are evaluated with jinja.
    A ``user`` var is available.
    """

    SUB: str | None = "{{ user.user_name }}"
    NAME: str | None = (
        "{% if user.formatted_name %}{{ user.formatted_name }}{% endif %}"
    )
    PHONE_NUMBER: str | None = (
        "{% if user.phone_numbers %}{{ user.phone_numbers[0] }}{% endif %}"
    )
    EMAIL: str | None = (
        "{% if user.preferred_email %}{{ user.preferred_email }}{% endif %}"
    )
    GIVEN_NAME: str | None = "{% if user.given_name %}{{ user.given_name }}{% endif %}"
    FAMILY_NAME: str | None = (
        "{% if user.family_name %}{{ user.family_name }}{% endif %}"
    )
    PREFERRED_USERNAME: str | None = (
        "{% if user.display_name %}{{ user.display_name }}{% endif %}"
    )
    LOCALE: str | None = (
        "{% if user.preferred_language %}{{ user.preferred_language }}{% endif %}"
    )
    ADDRESS: str | None = (
        "{% if user.formatted_address %}{{ user.formatted_address }}{% endif %}"
    )
    PICTURE: str | None = (
        "{% if user.photo %}{{ url_for('core.account.photo', user=user, field='photo', _external=True) }}{% endif %}"
    )
    WEBSITE: str | None = "{% if user.profile_url %}{{ user.profile_url }}{% endif %}"
    UPDATED_AT: str | None = (
        "{% if user.last_modified %}{{ user.last_modified.timestamp() }}{% endif %}"
    )


class OIDCSettings(BaseModel):
    """OpenID Connect settings.

    Belong in the ``CANAILLE_OIDC`` namespace.
    """

    ENABLE_OIDC: bool = True
    """Whether the Single Sign-On feature and the OpenID Connect API is enabled."""

    DYNAMIC_CLIENT_REGISTRATION_OPEN: bool = False
    """Whether a token is needed for the RFC7591 dynamical client registration.

    If :py:data:`True`, no token is needed to register a client.
    If :py:data:`False`, dynamical client registration needs a token defined in
    :attr:`DYNAMIC_CLIENT_REGISTRATION_TOKENS`.
    """

    DYNAMIC_CLIENT_REGISTRATION_TOKENS: list[str] | None = None
    """A list of tokens that can be used for dynamic client registration."""

    REQUIRE_NONCE: bool = True
    """Force the nonce exchange during the authentication flows.

    This adds security but may not be supported by all clients.
    """

    ACTIVE_JWKS: list[dict[str, str]] | None = None
    """The active JSON Web Keys Set.

    Those keys are used to sign and verify JWTs."""

    INACTIVE_JWKS: list[dict[str, str]] | None = None
    """The inactive JSON Web Keys Set.

    Those keys are only used to verify JWTs."""

    USERINFO_MAPPING: UserInfoMappingSettings | None = UserInfoMappingSettings()
    """"Attribute mapping used to build an OIDC UserInfo object.

    UserInfo is used to fill the id_token and the userinfo endpoint."""

    TRUSTED_DOMAINS: list[str] = [".localhost", "127.0.0.1"]
    """Trusted domains for automatic client trust.

    Clients with a :attr:`~canaille.oidc.basemodels.Client.client_uri` matching
    these domains will be automatically considered as trusted and will not
    display the consent page to users.
    This is particularly useful for development environments.

    Supports these patterns:

    - Exact match: ``example.com`` matches ``example.com``
    - Wildcard match: ``.example.com`` matches ``example.com`` and all its subdomains

    Examples:

    - ``[".localhost", "127.0.0.1"]`` (default for development)
    - ``[".dev.company.com", "staging.company.com"]``
    - ``[".local", "localhost"]``
    """



import datetime

from flask import Blueprint
from flask import abort
from flask import current_app
from flask import flash
from flask import g
from flask import redirect
from flask import request
from flask import url_for

from canaille.app import get_b64encoded_qr_image
from canaille.app.i18n import gettext as _
from canaille.app.otp import get_otp_authentication_setup_uri
from canaille.app.otp import is_otp_valid
from canaille.app.otp import make_otp_secret
from canaille.app.templating import render_template
from canaille.backends import Backend
from canaille.core.auth import auth_step
from canaille.core.auth import redirect_to_next_auth_step
from canaille.core.endpoints.forms import TwoFactorForm

bp = Blueprint("otp", __name__)


@bp.context_processor
def global_processor():
    return {
        "menu": False,
    }


@bp.route("/auth/otp", methods=["GET", "POST"])
@auth_step("otp")
def otp():
    if g.auth.user and not g.auth.user.last_otp_login:
        return redirect(url_for(".setup"))

    form = TwoFactorForm(request.form or None)
    form.render_field_macro_file = "core/partial/login_field.html"

    if not request.form or form.form_control():
        return render_template("core/auth/otp.html", form=form)

    if not form.validate() or not g.auth.user:
        flash(
            _("The passcode you entered is invalid. Please try again"),
            "error",
        )
        current_app.logger.security(f"Failed OTP authentication for {g.auth.user_name}")
        return render_template("core/auth/otp.html", form=form)

    otp_valid, hotp_counter = is_otp_valid(
        form.otp.data,
        current_app.config["CANAILLE"]["OTP_METHOD"],
        g.auth.user.secret_token,
        g.auth.user.hotp_counter,
    )
    if not otp_valid:
        flash(
            _("The passcode you entered is invalid. Please try again"),
            "error",
        )
        current_app.logger.security(f"Failed OTP authentication for {g.auth.user_name}")
        return render_template("core/auth/otp.html", form=form)

    g.auth.user.last_otp_login = datetime.datetime.now(datetime.timezone.utc)
    g.auth.user.hotp_counter = hotp_counter
    Backend.instance.save(g.auth.user)

    current_app.logger.security(f"Successful OTP authentication for {g.auth.user_name}")
    g.auth.set_step_finished("otp")
    return redirect_to_next_auth_step()


@bp.route("/auth/otp-setup", methods=["GET", "POST"])
def setup():
    if not current_app.features.has_otp:  # pragma: no cover
        abort(404)

    if not g.auth:
        flash(
            _("Cannot remember the login you attempted to sign in with."),
            "warning",
        )
        return redirect(
            url_for("core.account.profile_settings", edited_user=g.session.user)
            if g.session
            else url_for("core.auth.login")
        )

    if "otp_user_secret" not in g.auth.data:
        g.auth.data["otp_user_secret"] = make_otp_secret()

    form = TwoFactorForm(request.form or None)
    form.render_field_macro_file = "core/partial/login_field.html"
    uri, secret = get_otp_authentication_setup_uri(
        g.auth.user, g.auth.data["otp_user_secret"]
    )
    base64_qr_image = get_b64encoded_qr_image(uri)

    if not request.form or form.form_control():
        return render_template(
            "core/auth/setup-otp.html",
            secret=secret,
            qr_image=base64_qr_image,
            otp_uri=uri,
            form=form,
            menu=bool(g.session),
        )

    otp_valid, hotp_counter = is_otp_valid(
        form.otp.data,
        current_app.config["CANAILLE"]["OTP_METHOD"],
        g.auth.data["otp_user_secret"],
    )
    if not form.validate() or not otp_valid:
        flash(_("The passcode you entered is invalid. Please try again."), "error")
        return render_template(
            "core/auth/setup-otp.html",
            secret=secret,
            qr_image=base64_qr_image,
            otp_uri=uri,
            form=form,
            menu=bool(g.session),
        )

    g.auth.user.last_otp_login = datetime.datetime.now(datetime.timezone.utc)
    g.auth.user.secret_token = g.auth.data["otp_user_secret"]
    g.auth.user.hotp_counter = hotp_counter
    Backend.instance.save(g.auth.user)

    current_app.logger.security(
        f"OTP authentication factor reset for {g.auth.user.user_name}"
    )
    flash(_("Authenticator application correctly configured."), "success")

    g.auth.set_step_finished("otp")
    return redirect_to_next_auth_step()



from urllib.parse import parse_qs
from urllib.parse import urlsplit

from joserfc import jwt

from . import client_credentials


def test_request_object_query(testclient, logged_user, trusted_client, client_jwk):
    """Test an authentication request that uses the request parameter."""
    payload = dict(
        response_type="code",
        client_id=trusted_client.client_id,
        scope="openid profile email groups address phone",
        nonce="somenonce",
        redirect_uri=trusted_client.redirect_uris[0],
    )
    request_obj = jwt.encode({"alg": "RS256"}, payload, client_jwk)
    res = testclient.get(
        "/oauth/authorize",
        params=dict(client_id=trusted_client.client_id, request=request_obj),
        status=302,
    )

    assert res.location.startswith(trusted_client.redirect_uris[0])
    params = parse_qs(urlsplit(res.location).query)
    code = params["code"][0]

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="authorization_code",
            code=code,
            scope="openid profile email groups address phone",
            redirect_uri=trusted_client.redirect_uris[0],
        ),
        headers={"Authorization": f"Basic {client_credentials(trusted_client)}"},
        status=200,
    )

    access_token = res.json["access_token"]
    assert access_token


def test_request_object_uri_query(
    testclient, logged_user, trusted_client, client_jwk, httpserver
):
    """Test an authentication request that uses the request_uri parameter."""
    request_uri = f"http://{httpserver.host}:{httpserver.port}/request_obj"
    payload = dict(
        response_type="code",
        client_id=trusted_client.client_id,
        scope="openid profile email groups address phone",
        nonce="somenonce",
        redirect_uri=trusted_client.redirect_uris[0],
    )
    request_obj = jwt.encode({"alg": "RS256"}, payload, client_jwk)
    httpserver.expect_request("/request_obj").respond_with_data(request_obj)

    res = testclient.get(
        "/oauth/authorize",
        params=dict(client_id=trusted_client.client_id, request_uri=request_uri),
        status=302,
    )

    assert res.location.startswith(trusted_client.redirect_uris[0])
    params = parse_qs(urlsplit(res.location).query)
    code = params["code"][0]

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="authorization_code",
            code=code,
            scope="openid profile email groups address phone",
            redirect_uri=trusted_client.redirect_uris[0],
        ),
        headers={"Authorization": f"Basic {client_credentials(trusted_client)}"},
        status=200,
    )

    access_token = res.json["access_token"]
    assert access_token



import json
from unittest import mock

from canaille.commands import cli


def test_dump_stdout(cli_runner, backend, user, foo_group):
    """Test the full database dump command."""
    res = cli_runner.invoke(cli, ["dump"], catch_exceptions=False)
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "authorizationcode": [],
        "client": [],
        "consent": [],
        "group": [
            {
                "created": mock.ANY,
                "display_name": "foo",
                "id": foo_group.id,
                "last_modified": mock.ANY,
                "members": [
                    user.id,
                ],
            },
        ],
        "token": [],
        "user": [
            {
                "created": mock.ANY,
                "display_name": "Johnny",
                "emails": [
                    "john@doe.test",
                ],
                "family_name": "Doe",
                "formatted_address": "1234, some street, 6789 some city, some state",
                "formatted_name": "John (johnny) Doe",
                "given_name": "John",
                "groups": [foo_group.id],
                "id": user.id,
                "last_modified": mock.ANY,
                "password": mock.ANY,
                "phone_numbers": [
                    "555-000-000",
                ],
                "preferred_language": "en",
                "profile_url": "https://john.test",
                "user_name": "user",
                "locality": "some city",
                "postal_code": "6789",
                "region": "some state",
                "street": "1234, some street",
                "hotp_counter": 1,
                "last_otp_login": mock.ANY,
                "secret_token": mock.ANY,
            },
        ],
    }


def test_dump_stdout_explicit_model(cli_runner, backend, user, foo_group):
    """Test the full database dump command for a single model."""
    res = cli_runner.invoke(cli, ["dump", "user"], catch_exceptions=False)
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "user": [
            {
                "created": mock.ANY,
                "display_name": "Johnny",
                "emails": [
                    "john@doe.test",
                ],
                "family_name": "Doe",
                "formatted_address": "1234, some street, 6789 some city, some state",
                "formatted_name": "John (johnny) Doe",
                "given_name": "John",
                "groups": [foo_group.id],
                "id": user.id,
                "last_modified": mock.ANY,
                "password": mock.ANY,
                "phone_numbers": [
                    "555-000-000",
                ],
                "preferred_language": "en",
                "profile_url": "https://john.test",
                "user_name": "user",
                "locality": "some city",
                "postal_code": "6789",
                "region": "some state",
                "street": "1234, some street",
                "hotp_counter": 1,
                "last_otp_login": mock.ANY,
                "secret_token": mock.ANY,
            }
        ]
    }



import datetime
import json
import unittest.mock
from unittest import mock

from canaille.app import models
from canaille.backends import Backend
from canaille.commands import cli


def test_create(cli_runner, backend, foo_group):
    """Nominal case test for model create command."""
    res = cli_runner.invoke(
        cli,
        [
            "create",
            "user",
            "--formatted-name",
            "Johnny",
            "--emails",
            "foo@example.test",
            "--emails",
            "bar@example.test",
            "--given-name",
            "John",
            "--family-name",
            "Doe",
            "--user-name",
            "johnny",
            "--groups",
            foo_group.id,
            "--lock-date",
            "2050-01-01T10:10:10+00:00",
        ],
    )
    assert res.exit_code == 0, res.stdout
    output = json.loads(res.stdout)
    assert output == {
        "formatted_name": "Johnny",
        "created": mock.ANY,
        "last_modified": mock.ANY,
        "emails": [
            "foo@example.test",
            "bar@example.test",
        ],
        "family_name": "Doe",
        "given_name": "John",
        "id": mock.ANY,
        "user_name": "johnny",
        "groups": [foo_group.id],
        "lock_date": "2050-01-01T10:10:10+00:00",
    }
    user = backend.get(models.User, output["id"])
    backend.reload(foo_group)
    assert user.groups == [foo_group]
    assert user.lock_date == datetime.datetime(
        2050, 1, 1, 10, 10, 10, tzinfo=datetime.timezone.utc
    )
    backend.delete(user)


def test_create_quiet(cli_runner, backend, foo_group):
    """Test that --quiet produce no input."""
    res = cli_runner.invoke(
        cli,
        [
            "create",
            "user",
            "--quiet",
            "--formatted-name",
            "Johnny",
            "--emails",
            "foo@example.test",
            "--emails",
            "bar@example.test",
            "--given-name",
            "John",
            "--family-name",
            "Doe",
            "--user-name",
            "johnny",
            "--groups",
            foo_group.id,
            "--lock-date",
            "2050-01-01T10:10:10+00:00",
        ],
    )
    assert res.exit_code == 0, res.stdout
    assert res.stdout == ""
    user = backend.get(models.User, user_name="johnny")
    backend.reload(foo_group)
    assert user.groups == [foo_group]
    assert user.lock_date == datetime.datetime(
        2050, 1, 1, 10, 10, 10, tzinfo=datetime.timezone.utc
    )
    backend.delete(user)


def test_ignore_errors(cli_runner, backend, monkeypatch):
    """Test that error cases are ignored with --ignore-errors."""

    def mock_save(*args, **kwargs):
        raise Exception("Simulated save error")

    with unittest.mock.patch.object(Backend.instance, "save", side_effect=mock_save):
        res = cli_runner.invoke(
            cli,
            [
                "create",
                "user",
                "--formatted-name",
                "Test User",
                "--family-name",
                "User",
                "--user-name",
                "testuser",
            ],
        )
        assert res.exit_code != 0

    with unittest.mock.patch.object(Backend.instance, "save", side_effect=mock_save):
        res = cli_runner.invoke(
            cli,
            [
                "create",
                "user",
                "--ignore-errors",
                "--formatted-name",
                "Test User",
                "--family-name",
                "User",
                "--user-name",
                "testuser",
            ],
        )
        assert res.exit_code == 0
        assert res.stdout == ""



import inspect
import logging
import os
import uuid
from contextlib import contextmanager

import ldap.modlist
import ldif
from flask import current_app
from ldap.controls import DecodeControlTuples
from ldap.controls.ppolicy import PasswordPolicyControl
from ldap.controls.ppolicy import PasswordPolicyError
from ldap.controls.readentry import PostReadControl

from canaille.app import models
from canaille.app.configuration import CheckResult
from canaille.app.configuration import ConfigurationException
from canaille.app.i18n import gettext as _
from canaille.app.models import MODELS
from canaille.backends import Backend
from canaille.backends import ModelEncoder
from canaille.backends import get_lockout_delay_message
from canaille.backends import is_meaningful_value
from canaille.backends.models import Model

from .utils import listify
from .utils import python_attrs_to_ldap


@contextmanager
def ldap_connection(config):
    conn = ldap.initialize(config["CANAILLE_LDAP"]["URI"])
    conn.set_option(ldap.OPT_NETWORK_TIMEOUT, config["CANAILLE_LDAP"]["TIMEOUT"])
    conn.simple_bind_s(
        config["CANAILLE_LDAP"]["BIND_DN"], config["CANAILLE_LDAP"]["BIND_PW"]
    )

    try:
        yield conn
    finally:
        conn.unbind_s()


def install_schema(config, schema_path):
    from canaille.app.installation import InstallationException

    with open(schema_path) as fd:
        parser = ldif.LDIFRecordList(fd)
        parser.parse()

    try:
        with ldap_connection(config) as conn:
            for dn, entry in parser.all_records:
                add_modlist = ldap.modlist.addModlist(entry)
                conn.add_s(dn, add_modlist)

    except ldap.INSUFFICIENT_ACCESS as exc:
        raise InstallationException(
            f"The user '{config['CANAILLE_LDAP']['BIND_DN']}' has insufficient permissions to install LDAP schemas."
        ) from exc


class LDAPModelEncoder(ModelEncoder):
    def sanitize_attr(self, obj, attr):
        """Replace dns by uuids in model attributes."""
        value = getattr(obj, attr)
        type_, _ = obj.get_model_annotations(attr)
        is_model = inspect.isclass(type_) and issubclass(type_, Model)
        if not is_model:
            return value

        if isinstance(value, list):
            return [item.id for item in value]

        return value.id

    def default(self, obj):
        if isinstance(obj, Model):
            sanitized_state = {
                attr: self.sanitize_attr(obj, attr) for attr in obj.attributes
            }
            return {
                key: value
                for key, value in sanitized_state.items()
                if is_meaningful_value(value)
            }
        return super().default(obj)


class LDAPBackend(Backend):
    json_encoder = LDAPModelEncoder

    def __init__(self, config):
        super().__init__(config)
        self._connection = None
        setup_ldap_models(config)

    @classmethod
    def install(cls, app):
        cls.setup_schemas(app.config)
        with cls(app.config).session():
            models.Token.install()
            models.AuthorizationCode.install()
            models.Client.install()
            models.Consent.install()

    @classmethod
    def setup_schemas(cls, config):
        from .ldapobject import LDAPObject

        with cls(config).session():
            if "oauthClient" not in LDAPObject.ldap_object_classes(force=True):
                install_schema(
                    config,
                    os.path.dirname(__file__) + "/schemas/oauth2-openldap.ldif",
                )

    @property
    def connection(self):
        if self._connection:
            return self._connection

        try:
            self._connection = ldap.initialize(self.config["CANAILLE_LDAP"]["URI"])
            self._connection.set_option(
                ldap.OPT_NETWORK_TIMEOUT,
                self.config["CANAILLE_LDAP"]["TIMEOUT"],
            )
            self._connection.simple_bind_s(
                self.config["CANAILLE_LDAP"]["BIND_DN"],
                self.config["CANAILLE_LDAP"]["BIND_PW"],
            )

        except ldap.SERVER_DOWN as exc:
            message = _("Could not connect to the LDAP server '{uri}'").format(
                uri=self.config["CANAILLE_LDAP"]["URI"]
            )
            logging.error(message)
            raise ConfigurationException(message) from exc

        except ldap.INVALID_CREDENTIALS as exc:
            message = _("LDAP authentication failed with user '{user}'").format(
                user=self.config["CANAILLE_LDAP"]["BIND_DN"]
            )
            logging.error(message)
            raise ConfigurationException(message) from exc

        return self._connection

    def teardown(self):
        if self._connection:  # pragma: no branch
            self._connection.unbind_s()
            self._connection = None

    @classmethod
    def check_network_config(cls, config):
        from canaille.app import models

        with cls(config).session():
            try:
                user = models.User(
                    formatted_name=f"canaille_{uuid.uuid4()}",
                    family_name=f"canaille_{uuid.uuid4()}",
                    user_name=f"canaille_{uuid.uuid4()}",
                    emails=f"canaille_{uuid.uuid4()}@mydomain.example",
                    password="correct horse battery staple",
                )
                Backend.instance.save(user)
                Backend.instance.delete(user)

            except ldap.INSUFFICIENT_ACCESS:
                return CheckResult(
                    message=f"LDAP user '{config['CANAILLE_LDAP']['BIND_DN']}' cannot create "
                    f"users at '{config['CANAILLE_LDAP']['USER_BASE']}'",
                    success=False,
                )

            try:
                models.Group.ldap_object_classes()

                user = models.User(
                    cn=f"canaille_{uuid.uuid4()}",
                    family_name=f"canaille_{uuid.uuid4()}",
                    user_name=f"canaille_{uuid.uuid4()}",
                    emails=f"canaille_{uuid.uuid4()}@mydomain.example",
                    password="correct horse battery staple",
                )
                Backend.instance.save(user)

                group = models.Group(
                    display_name=f"canaille_{uuid.uuid4()}",
                    members=[user],
                )
                Backend.instance.save(group)
                Backend.instance.delete(group)

            except ldap.INSUFFICIENT_ACCESS:
                return CheckResult(
                    message=f"LDAP user '{config['CANAILLE_LDAP']['BIND_DN']}' cannot create "
                    f"groups at '{config['CANAILLE_LDAP']['GROUP_BASE']}'",
                    success=False,
                )

            finally:
                Backend.instance.delete(user)

        return CheckResult(
            message="The connection to the LDAP server and permissions of the bind DN are correct",
            success=True,
        )

    def has_account_lockability(self):
        from .ldapobject import LDAPObject

        try:
            return "pwdEndTime" in LDAPObject.ldap_object_attributes()
        except ldap.SERVER_DOWN:  # pragma: no cover
            return False

    def check_user_password(self, user, password):
        if current_app.features.has_intruder_lockout:
            if current_lockout_delay := user.get_intruder_lockout_delay():
                self.save(user)
                return (False, get_lockout_delay_message(current_lockout_delay))

        conn = ldap.initialize(current_app.config["CANAILLE_LDAP"]["URI"])

        conn.set_option(
            ldap.OPT_NETWORK_TIMEOUT,
            current_app.config["CANAILLE_LDAP"]["TIMEOUT"],
        )

        message = None
        try:
            res = conn.simple_bind_s(
                user.dn, password, serverctrls=[PasswordPolicyControl()]
            )
            controls = res[3]
            result = True
        except ldap.INVALID_CREDENTIALS as exc:
            controls = DecodeControlTuples(exc.args[0]["ctrls"])
            result = False
        finally:
            conn.unbind_s()

        for control in controls:

            def gettext(x):
                return x

            if (
                control.controlType == PasswordPolicyControl.controlType
                and control.error == PasswordPolicyError.namedValues["accountLocked"]
            ):
                message = gettext("Your account has been locked.")
            elif (
                control.controlType == PasswordPolicyControl.controlType
                and control.error == PasswordPolicyError.namedValues["changeAfterReset"]
            ):
                message = gettext("You should change your password.")

        return result, message

    def set_user_password(self, user, password):
        conn = self.connection
        conn.passwd_s(
            user.dn,
            None,
            password.encode("utf-8"),
        )

    def do_query(self, model, dn=None, filter=None, *args, **kwargs):
        from .ldapobjectquery import LDAPObjectQuery

        base = dn
        if dn is None:
            base = f"{model.base},{model.root_dn}"
        elif "=" not in base:
            base = ldap.dn.escape_dn_chars(base)
            base = f"{model.rdn_attribute}={base},{model.base},{model.root_dn}"

        class_filter = (
            "".join([f"(objectClass={oc})" for oc in model.ldap_object_class])
            if model.ldap_object_class
            else ""
        )
        if class_filter:
            class_filter = f"(|{class_filter})"

        arg_filter = ""
        ldap_args = python_attrs_to_ldap(
            {
                model.python_attribute_to_ldap(name): values
                for name, values in kwargs.items()
                if values is not None
            },
            encode=False,
        )
        for key, value in ldap_args.items():
            if len(value) == 1:
                escaped_value = ldap.filter.escape_filter_chars(value[0])
                arg_filter += f"({key}={escaped_value})"

            else:
                values = [ldap.filter.escape_filter_chars(v) for v in value]
                arg_filter += (
                    "(|" + "".join([f"({key}={value})" for value in values]) + ")"
                )

        if not filter:
            filter = ""

        ldapfilter = f"(&{class_filter}{arg_filter}{filter})"
        base = base or f"{model.base},{model.root_dn}"
        try:
            result = self.connection.search_s(
                base, ldap.SCOPE_SUBTREE, ldapfilter or None, ["+", "*"]
            )
        except ldap.NO_SUCH_OBJECT:
            result = []
        return LDAPObjectQuery(model, result)

    def fuzzy(self, model, query, attributes=None, **kwargs):
        query = ldap.filter.escape_filter_chars(query)
        attributes = attributes or model.may() + model.must()
        attributes = [model.python_attribute_to_ldap(name) for name in attributes]
        filter = (
            "(|" + "".join(f"({attribute}=*{query}*)" for attribute in attributes) + ")"
        )
        return self.query(model, filter=filter, **kwargs)

    def get(self, model, identifier=None, /, **kwargs):
        try:
            return self.query(model, identifier, **kwargs)[0]
        except (IndexError, ldap.NO_SUCH_OBJECT):
            if identifier and model.base:
                return (
                    self.get(model, **{model.identifier_attribute: identifier})
                    or self.get(model, id=identifier)
                    or None
                )

            return None

    def do_restore(self, payload):
        # Canaille exports uuids but LDAP uuids are read-only
        # thus uuids will probably be replaced by LDAP.
        # However as they server as identifiers in the dumps,
        # we need to keep a temporary uuid <-> dn mapping
        uuids_dn = {}

        # Create models without attributes that are references to other models
        # except when those attributes are required
        # We *could* do this in one single loop, but Client.audience is self-referencing,
        # so it has to be created first, and then must be added to its own audience.
        for model_name, model in MODELS.items():
            states = payload.get(model_name, [])
            for state in states:
                filtered_state = self.filter_state(
                    model, state, keep_non_model_and_required_attrs=True
                )
                filtered_state = self.replace_uuids_with_dns(
                    model, filtered_state, uuids_dn
                )
                obj = model(**filtered_state)
                Backend.instance.save(obj)
                uuids_dn[state["id"]] = obj.dn

        # Insert attributes that are model references
        for model_name, model in MODELS.items():
            states = payload.get(model_name, [])
            for state in states:
                obj_dn = uuids_dn[state["id"]]
                filtered_state = self.filter_state(
                    model,
                    state,
                    keep_non_model_and_required_attrs=False,
                )
                filtered_state = self.replace_uuids_with_dns(
                    model, filtered_state, uuids_dn
                )
                obj = Backend.instance.get(model, obj_dn)
                for attr, value in filtered_state.items():
                    setattr(obj, attr, value)
                Backend.instance.save(obj)

    @classmethod
    def filter_state(cls, model, state, keep_non_model_and_required_attrs):
        def filter_attribute(attr):
            type_, _ = model.get_model_annotations(attr)
            is_non_model = not inspect.isclass(type_) or not issubclass(type_, Model)
            is_required = model.is_attr_required(attr)
            is_writable = not model.is_attr_readonly(attr)
            valid = (
                is_writable
                and (is_non_model or is_required) == keep_non_model_and_required_attrs
            )
            return valid

        return {attr: value for attr, value in state.items() if filter_attribute(attr)}

    @classmethod
    def replace_uuids_with_dns(cls, model, state, uuids_dn):
        def get_dn_from_uuid(uuid, model):
            backend_model = MODELS.get(model.__name__.lower())
            return uuids_dn.get(uuid) or Backend.instance.get(backend_model, id=uuid).dn

        def replace_attr(attr, value):
            type_, _ = model.get_model_annotations(attr)
            is_model = inspect.isclass(type_) and issubclass(type_, Model)
            if not is_model:
                return value

            if isinstance(value, list):
                return [get_dn_from_uuid(item, type_) for item in value]

            return get_dn_from_uuid(value, type_)

        return {attr: replace_attr(attr, value) for attr, value in state.items()}

    def do_save(self, instance):
        current_object_classes = instance.get_ldap_attribute("objectClass") or []
        instance.set_ldap_attribute(
            "objectClass",
            list(set(instance.ldap_object_class) | set(current_object_classes)),
        )

        # PostReadControl allows to read the updated object attributes on creation/edition
        attributes = ["objectClass"] + [
            instance.python_attribute_to_ldap(name) for name in instance.attributes
        ]
        read_post_control = PostReadControl(criticality=True, attrList=attributes)

        # Object already exists in the LDAP database
        if instance.exists:
            deletions = [
                name
                for name, value in instance.changes.items()
                if (
                    not is_meaningful_value(value)
                    or (
                        isinstance(value, list)
                        and len(value) == 1
                        and not is_meaningful_value(value[0])
                    )
                )
                and name in instance.state
            ]
            changes = {
                name: value
                for name, value in instance.changes.items()
                if name not in deletions and instance.state.get(name) != value
            }
            formatted_changes = python_attrs_to_ldap(changes, null_allowed=False)
            modlist = [(ldap.MOD_DELETE, name, None) for name in deletions] + [
                (ldap.MOD_REPLACE, name, values)
                for name, values in formatted_changes.items()
            ]
            _, _, _, [result] = self.connection.modify_ext_s(
                instance.dn, modlist, serverctrls=[read_post_control]
            )

        # Object does not exist yet in the LDAP database
        else:
            changes = {
                name: value
                for name, value in {**instance.state, **instance.changes}.items()
                if value and is_meaningful_value(value[0])
            }
            formatted_changes = python_attrs_to_ldap(changes, null_allowed=False)
            modlist = [(name, values) for name, values in formatted_changes.items()]
            _, _, _, [result] = self.connection.add_ext_s(
                instance.dn, modlist, serverctrls=[read_post_control]
            )

        instance.exists = True
        instance.state = {**result.entry, **instance.changes}
        instance.changes = {}

    def do_delete(self, instance):
        try:
            self.connection.delete_s(instance.dn)
        except ldap.NO_SUCH_OBJECT:
            pass

    def do_reload(self, instance):
        result = self.connection.search_s(
            instance.dn, ldap.SCOPE_SUBTREE, None, ["+", "*"]
        )
        instance.changes = {}
        instance.state = result[0][1]


def setup_ldap_models(config):
    from canaille.app import models

    from .ldapobject import LDAPObject

    LDAPObject.root_dn = config["CANAILLE_LDAP"]["ROOT_DN"]

    user_base = config["CANAILLE_LDAP"]["USER_BASE"].replace(
        f",{config['CANAILLE_LDAP']['ROOT_DN']}", ""
    )
    models.User.base = user_base
    models.User.rdn_attribute = config["CANAILLE_LDAP"]["USER_RDN"]
    object_class = config["CANAILLE_LDAP"]["USER_CLASS"]
    models.User.ldap_object_class = listify(object_class)

    group_base = config["CANAILLE_LDAP"]["GROUP_BASE"].replace(
        f",{config['CANAILLE_LDAP']['ROOT_DN']}", ""
    )
    models.Group.base = group_base or None
    models.Group.rdn_attribute = config["CANAILLE_LDAP"]["GROUP_RDN"]
    object_class = config["CANAILLE_LDAP"]["GROUP_CLASS"]
    models.Group.ldap_object_class = listify(object_class)



import click
from flask.cli import with_appcontext

from canaille.app.commands import with_backendcontext

try:
    HAS_FAKER = True
except ImportError:
    HAS_FAKER = False


@click.group()
@click.option("--nb", default=1, help="Number of items to create")
@click.pass_context
@with_appcontext
def populate(ctx, nb):
    """Populate the database with generated random data."""
    ctx.ensure_object(dict)

    ctx.obj["number"] = nb


@populate.command()
@click.pass_context
@with_appcontext
@with_backendcontext
def users(ctx):
    """Populate the database with generated random users."""
    from canaille.core.populate import fake_users

    fake_users(ctx.obj["number"])


@populate.command()
@click.pass_context
@click.option(
    "--nb-users-max",
    default=1,
    help="The maximum number of users that will randomly be affected in the group",
)
@with_appcontext
@with_backendcontext
def groups(ctx, nb_users_max):
    """Populate the database with generated random groups."""
    from canaille.core.populate import fake_groups

    fake_groups(ctx.obj["number"], nb_users_max)


def register(cli):
    if HAS_FAKER:  # pragma: no branch
        cli.add_command(populate)



import logging

import pytest
from flask import g
from webtest import Upload

from canaille.core.populate import fake_users


@pytest.fixture
def configuration(configuration):
    configuration["CANAILLE"]["EMAIL_CONFIRMATION"] = False
    return configuration


def test_invalid_form_request(testclient, logged_user):
    res = testclient.get("/profile/user")
    res = res.forms["baseform"].submit(
        name="action", value="invalid-action", status=400
    )


def test_user_list_pagination(testclient, logged_admin, backend):
    res = testclient.get("/users")
    res.mustcontain("1 item")

    users = fake_users(25)

    res = testclient.get("/users")
    res.mustcontain("26 items")
    user_name = res.pyquery(".users tbody tr:nth-of-type(1) td:nth-of-type(2) a").text()
    assert user_name

    form = res.forms["tableform"]
    res = form.submit(name="page", value="2")
    assert user_name not in res.pyquery(
        ".users tbody tr td:nth-of-type(2) a"
    ).text().split(" ")
    for user in users:
        backend.delete(user)

    res = testclient.get("/users")
    res.mustcontain("1 item")


def test_user_list_bad_pages(testclient, logged_admin):
    res = testclient.get("/users")
    form = res.forms["tableform"]
    testclient.post(
        "/users", {"csrf_token": form["csrf_token"].value, "page": "2"}, status=404
    )

    res = testclient.get("/users")
    form = res.forms["tableform"]
    testclient.post(
        "/users", {"csrf_token": form["csrf_token"].value, "page": "-1"}, status=404
    )


def test_user_list_search(testclient, logged_admin, user, moderator):
    res = testclient.get("/users")
    res.mustcontain("3 items")
    res.mustcontain(moderator.formatted_name)
    res.mustcontain(user.formatted_name)

    form = res.forms["search"]
    form["query"] = "Jack"
    res = form.submit()

    res.mustcontain("1 item")
    res.mustcontain(moderator.formatted_name)
    res.mustcontain(no=user.formatted_name)


def test_user_list_search_only_allowed_fields(
    testclient, logged_admin, user, moderator, backend
):
    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["READ"].remove("user_name")
    testclient.app.config["CANAILLE"]["ACL"]["ADMIN"]["READ"].remove("user_name")

    res = testclient.get("/users")
    form = res.forms["search"]
    form["query"] = "user"
    res = form.submit()

    res.mustcontain("0 items")
    res.mustcontain(no=user.formatted_name)
    res.mustcontain(no=moderator.formatted_name)


def test_edition_permission(
    testclient,
    logged_user,
    admin,
    backend,
):
    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["PERMISSIONS"] = []
    backend.reload(logged_user)
    testclient.get("/profile/user", status=404)

    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["PERMISSIONS"] = ["edit_self"]
    backend.reload(g.session.user)
    testclient.get("/profile/user", status=200)


def test_edition(testclient, logged_user, admin, jpeg_photo, backend, caplog):
    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    form["given_name"] = "given_name"
    form["family_name"] = "family_name"
    form["display_name"] = "display_name"
    form["emails-0"] = "email@mydomain.test"
    form["phone_numbers-0"] = "555-666-777"
    form["formatted_address"] = "formatted_address"
    form["street"] = "street"
    form["postal_code"] = "postal_code"
    form["locality"] = "locality"
    form["region"] = "region"
    form["employee_number"] = 666
    form["department"] = 1337
    form["title"] = "title"
    form["organization"] = "organization"
    form["preferred_language"] = "fr"
    form["photo"] = Upload("logo.jpg", jpeg_photo)

    res = form.submit(name="action", value="edit-profile")
    assert res.flashes == [("success", "Le profil a été mis à jour avec succès.")], (
        res.text
    )
    assert (
        "canaille",
        logging.SECURITY,
        "Updated email for user",
    ) in caplog.record_tuples
    res = res.follow()

    backend.reload(logged_user)

    assert logged_user.given_name == "given_name"
    assert logged_user.family_name == "family_name"
    assert logged_user.display_name == "display_name"
    assert logged_user.emails == ["email@mydomain.test"]
    assert logged_user.phone_numbers == ["555-666-777"]
    assert logged_user.formatted_address == "formatted_address"
    assert logged_user.street == "street"
    assert logged_user.postal_code == "postal_code"
    assert logged_user.locality == "locality"
    assert logged_user.region == "region"
    assert logged_user.preferred_language == "fr"
    assert logged_user.employee_number == "666"
    assert logged_user.department == "1337"
    assert logged_user.title == "title"
    assert logged_user.organization == "organization"
    assert logged_user.photo == jpeg_photo

    logged_user.formatted_name = "John (johnny) Doe"
    logged_user.family_name = "Doe"
    logged_user.emails = ["john@doe.test"]
    logged_user.given_name = None
    logged_user.photo = None
    backend.save(logged_user)


def test_edition_remove_fields(
    testclient,
    logged_user,
    admin,
    backend,
):
    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    form["display_name"] = ""
    form["phone_numbers-0"] = ""

    res = form.submit(name="action", value="edit-profile")
    assert res.flashes == [("success", "Profile updated successfully.")], res.text
    res = res.follow()

    backend.reload(logged_user)

    assert not logged_user.display_name
    assert not logged_user.phone_numbers

    logged_user.formatted_name = "John (johnny) Doe"
    logged_user.family_name = "Doe"
    logged_user.emails = ["john@doe.test"]
    logged_user.given_name = None
    logged_user.photo = None
    backend.save(logged_user)


def test_field_permissions_none(testclient, logged_user, backend):
    logged_user.phone_numbers = ["555-666-777"]
    backend.save(logged_user)

    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"] = {
        "READ": ["user_name"],
        "WRITE": [],
        "PERMISSIONS": ["edit_self"],
        "FILTER": None,
    }

    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    assert "phone_numbers-0" not in form.fields

    res = testclient.post(
        "/profile/user",
        {
            "action": "edit-profile",
            "phone_numbers-0": "000-000-000",
            "csrf_token": form["csrf_token"].value,
        },
    )
    assert ("success", "Profile updated successfully.") in res.flashes

    backend.reload(logged_user)
    assert logged_user.phone_numbers == ["555-666-777"]


def test_field_permissions_read(testclient, logged_user, backend):
    logged_user.phone_numbers = ["555-666-777"]
    backend.save(logged_user)

    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"] = {
        "READ": ["user_name", "phone_numbers"],
        "WRITE": [],
        "PERMISSIONS": ["edit_self"],
        "FILTER": None,
    }

    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    assert "phone_numbers-0" in form.fields

    res = testclient.post(
        "/profile/user",
        {
            "action": "edit-profile",
            "phone_numbers-0": "000-000-000",
            "csrf_token": form["csrf_token"].value,
        },
    )
    assert ("error", "Profile edition failed.") in res.flashes

    backend.reload(logged_user)
    assert logged_user.phone_numbers == ["555-666-777"]


def test_field_permissions_write(testclient, logged_user, backend):
    logged_user.phone_numbers = ["555-666-777"]
    backend.save(logged_user)

    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"] = {
        "READ": ["user_name"],
        "WRITE": ["phone_numbers"],
        "PERMISSIONS": ["edit_self"],
        "FILTER": None,
    }

    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    assert "phone_numbers-0" in form.fields

    res = testclient.post(
        "/profile/user",
        {
            "action": "edit-profile",
            "phone_numbers-0": "000-000-000",
            "csrf_token": form["csrf_token"].value,
        },
    )
    assert ("success", "Profile updated successfully.") in res.flashes

    backend.reload(logged_user)
    assert logged_user.phone_numbers == ["000-000-000"]


def test_simple_user_cannot_edit_other(testclient, admin, logged_user):
    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    testclient.get("/profile/admin", status=404)
    testclient.post(
        "/profile/admin",
        {"action": "edit-profile", "csrf_token": form["csrf_token"].value},
        status=404,
    )
    testclient.post(
        "/profile/admin",
        {"action": "delete", "csrf_token": form["csrf_token"].value},
        status=404,
    )
    testclient.get("/users", status=403)


def test_admin_bad_request(testclient, logged_moderator):
    testclient.get("/profile/foobar", status=404)


def test_bad_email(testclient, logged_user, backend):
    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]

    form["emails-0"] = "john@doe.test"

    res = form.submit(name="action", value="edit-profile").follow()

    assert ["john@doe.test"] == logged_user.emails

    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]

    form["emails-0"] = "yolo"

    res = form.submit(name="action", value="edit-profile", status=200)

    backend.reload(logged_user)

    assert ["john@doe.test"] == logged_user.emails


def test_surname_is_mandatory(testclient, logged_user, backend):
    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    logged_user.family_name = "Doe"

    form["family_name"] = ""

    res = form.submit(name="action", value="edit-profile", status=200)

    backend.reload(logged_user)

    assert "Doe" == logged_user.family_name


def test_formcontrol(testclient, logged_user):
    res = testclient.get("/profile/user")
    form = res.forms["baseform"]
    assert "emails-1" not in form.fields

    res = form.submit(status=200, name="fieldlist_add", value="emails-0")
    form = res.forms["baseform"]
    assert "emails-1" in form.fields


def test_formcontrol_htmx(testclient, logged_user):
    res = testclient.get("/profile/user")
    form = res.forms["baseform"]
    data = {
        field: form[field].value
        for field in form.fields
        if len(form.fields.get(field)) == 1
    }
    data["fieldlist_add"] = "emails-0"
    response = testclient.post(
        "/profile/user",
        data,
        headers={
            "HX-Request": "true",
            "HX-Trigger-Name": "listfield_add",
        },
    )
    assert "emails-0" in response.text
    assert "emails-1" in response.text


def test_inline_validation(testclient, logged_admin, user):
    res = testclient.get("/profile/admin")
    form = res.forms["baseform"]
    res = testclient.post(
        "/profile/admin",
        {
            "csrf_token": form["csrf_token"].value,
            "emails-0": "john@doe.test",
            "action": "edit-profile",
        },
        headers={
            "HX-Request": "true",
            "HX-Trigger-Name": "emails-0",
        },
    )
    res.mustcontain("The email 'john@doe.test' is already used")


def test_inline_validation_keep_indicators(testclient, logged_admin, user, backend):
    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["WRITE"].remove("display_name")
    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["READ"].append("display_name")
    testclient.app.config["CANAILLE"]["ACL"]["ADMIN"]["WRITE"].append("display_name")
    backend.reload(logged_admin)
    backend.reload(user)

    res = testclient.get("/profile/admin")
    form = res.forms["baseform"]
    res = testclient.post(
        "/profile/user",
        {
            "csrf_token": form["csrf_token"].value,
            "display_name": "George Abitbol",
            "action": "edit-profile",
        },
        headers={
            "HX-Request": "true",
            "HX-Trigger-Name": "display_name",
        },
    )
    res.mustcontain("This user cannot edit this field")



import datetime
from urllib.parse import parse_qs
from urllib.parse import urlsplit

import time_machine

from canaille.app import models

from . import client_credentials


def test_nominal_case(testclient, user, client, trusted_client, backend):
    max_age = int(datetime.timedelta(hours=3).total_seconds())

    with time_machine.travel("2025-01-01 02:30:00+00:00", tick=False):
        res = testclient.get("/login")

        res.form["login"] = "user"
        res = res.form.submit(name="answer", value="accept", status=302).follow()

        res.form["password"] = "correct horse battery staple"
        res = res.form.submit(name="answer", value="accept", status=302).follow()

    with time_machine.travel("2025-01-01 03:00:00+00:00", tick=False):
        res = testclient.get(
            "/oauth/authorize",
            params=dict(
                response_type="code",
                client_id=client.client_id,
                scope="openid",
                nonce="somenonce",
                redirect_uri="https://client.test/redirect1",
                max_age=max_age,
            ),
        )
        res = res.form.submit(name="answer", value="accept", status=302)

    params = parse_qs(urlsplit(res.location).query)
    code = params["code"][0]

    with time_machine.travel("2025-01-01 03:01:00+00:00", tick=False):
        res = testclient.post(
            "/oauth/token",
            params=dict(
                grant_type="authorization_code",
                code=code,
                scope="openid",
                redirect_uri=client.redirect_uris[0],
            ),
            headers={"Authorization": f"Basic {client_credentials(client)}"},
        )

    for consent in backend.query(models.Consent, client=client, subject=user):
        backend.delete(consent)


def test_max_age_expired(testclient, user, client, trusted_client, backend):
    max_age = int(datetime.timedelta(minutes=15).total_seconds())

    with time_machine.travel("2025-01-01 02:30:00+00:00", tick=False):
        res = testclient.get("/login")

        res.form["login"] = "user"
        res = res.form.submit(name="answer", value="accept", status=302).follow()

        res.form["password"] = "correct horse battery staple"
        res = res.form.submit(name="answer", value="accept", status=302).follow()

    with time_machine.travel("2025-01-01 03:00:00+00:00", tick=False):
        res = testclient.get(
            "/oauth/authorize",
            params=dict(
                response_type="code",
                client_id=client.client_id,
                scope="openid",
                nonce="somenonce",
                redirect_uri="https://client.test/redirect1",
                max_age=max_age,
            ),
        ).follow()

        res.form["password"] = "correct horse battery staple"
        res = res.form.submit(name="answer", value="accept", status=302).follow()

        res = res.form.submit(name="answer", value="accept", status=302)

    params = parse_qs(urlsplit(res.location).query)
    code = params["code"][0]

    with time_machine.travel("2025-01-01 03:01:00+00:00", tick=False):
        res = testclient.post(
            "/oauth/token",
            params=dict(
                grant_type="authorization_code",
                code=code,
                scope="openid",
                redirect_uri=client.redirect_uris[0],
            ),
            headers={"Authorization": f"Basic {client_credentials(client)}"},
        )

    for consent in backend.query(models.Consent, client=client, subject=user):
        backend.delete(consent)



def test_content_security_policy(testclient):
    res = testclient.get("/login", status=200)
    assert (
        res.headers["Content-Security-Policy"]
        == "default-src 'self'; font-src 'self' data:; img-src 'self' data: https:"
    )



from flask import g


def test_oauth_authorization_server(testclient):
    res = testclient.get("/.well-known/oauth-authorization-server", status=200).json
    assert "http://canaille.test" == res["issuer"]
    assert res == {
        "authorization_endpoint": "http://canaille.test/oauth/authorize",
        "code_challenge_methods_supported": ["plain", "S256"],
        "grant_types_supported": [
            "authorization_code",
            "implicit",
            "password",
            "client_credentials",
            "refresh_token",
        ],
        "introspection_endpoint": "http://canaille.test/oauth/introspect",
        "introspection_endpoint_auth_methods_supported": ["client_secret_basic"],
        "issuer": "http://canaille.test",
        "jwks_uri": "http://canaille.test/oauth/jwks.json",
        "registration_endpoint": "http://canaille.test/oauth/register",
        "response_types_supported": [
            "code",
            "token",
            "id_token",
            "code token",
            "code id_token",
            "token id_token",
            "code id_token token",
        ],
        "revocation_endpoint": "http://canaille.test/oauth/revoke",
        "revocation_endpoint_auth_methods_supported": [
            "client_secret_basic",
        ],
        "scopes_supported": [
            "openid",
            "profile",
            "email",
            "address",
            "phone",
            "groups",
        ],
        "service_documentation": "https://canaille.readthedocs.io",
        "token_endpoint": "http://canaille.test/oauth/token",
        "token_endpoint_auth_methods_supported": [
            "client_secret_basic",
            "private_key_jwt",
            "client_secret_post",
            "none",
        ],
        "token_endpoint_auth_signing_alg_values_supported": [
            "RS256",
            "ES256",
            "HS256",
            "EdDSA",
        ],
        "ui_locales_supported": g.available_language_codes,
        "userinfo_endpoint": "http://canaille.test/oauth/userinfo",
        "authorization_response_iss_parameter_supported": True,
    }


def test_openid_configuration(testclient):
    res = testclient.get("/.well-known/openid-configuration", status=200).json
    assert "http://canaille.test" == res["issuer"]
    assert res == {
        "authorization_endpoint": "http://canaille.test/oauth/authorize",
        "claim_types_supported": [
            "normal",
        ],
        "claims_parameter_supported": False,
        "claims_supported": [
            "sub",
            "iss",
            "auth_time",
            "acr",
            "name",
            "given_name",
            "family_name",
            "nickname",
            "profile",
            "picture",
            "website",
            "email",
            "email_verified",
            "locale",
            "zoneinfo",
            "groups",
            "nonce",
        ],
        "code_challenge_methods_supported": ["plain", "S256"],
        "end_session_endpoint": "http://canaille.test/oauth/end_session",
        "grant_types_supported": [
            "authorization_code",
            "implicit",
            "password",
            "client_credentials",
            "refresh_token",
        ],
        "id_token_signing_alg_values_supported": [
            "none",
            "RS256",
            "ES256",
            "HS256",
            "EdDSA",
        ],
        "userinfo_signing_alg_values_supported": [
            "none",
            "RS256",
            "ES256",
            "HS256",
            "EdDSA",
        ],
        "introspection_endpoint": "http://canaille.test/oauth/introspect",
        "introspection_endpoint_auth_methods_supported": ["client_secret_basic"],
        "issuer": "http://canaille.test",
        "jwks_uri": "http://canaille.test/oauth/jwks.json",
        "registration_endpoint": "http://canaille.test/oauth/register",
        "request_parameter_supported": True,
        "request_uri_parameter_supported": True,
        "require_request_uri_registration": False,
        "response_modes_supported": [
            "query",
            "fragment",
        ],
        "response_types_supported": [
            "code",
            "token",
            "id_token",
            "code token",
            "code id_token",
            "token id_token",
            "code id_token token",
        ],
        "revocation_endpoint": "http://canaille.test/oauth/revoke",
        "revocation_endpoint_auth_methods_supported": [
            "client_secret_basic",
        ],
        "scopes_supported": [
            "openid",
            "profile",
            "email",
            "address",
            "phone",
            "groups",
        ],
        "service_documentation": "https://canaille.readthedocs.io",
        "subject_types_supported": ["public"],
        "token_endpoint": "http://canaille.test/oauth/token",
        "token_endpoint_auth_methods_supported": [
            "client_secret_basic",
            "private_key_jwt",
            "client_secret_post",
            "none",
        ],
        "token_endpoint_auth_signing_alg_values_supported": [
            "RS256",
            "ES256",
            "HS256",
            "EdDSA",
        ],
        "ui_locales_supported": g.available_language_codes,
        "userinfo_endpoint": "http://canaille.test/oauth/userinfo",
        "prompt_values_supported": ["none", "login", "consent"],
        "authorization_response_iss_parameter_supported": True,
        "request_object_signing_alg_values_supported": [
            "none",
            "RS256",
            "ES256",
            "HS256",
            "EdDSA",
        ],
    }


def test_openid_configuration_prompt_value_create(testclient):
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    res = testclient.get("/.well-known/openid-configuration", status=200).json
    assert "create" in res["prompt_values_supported"]



from canaille.commands import cli


def test_install(cli_runner):
    """Tests that the install command is successfull.

    More detailed tests should be run by backends.
    """
    res = cli_runner.invoke(cli, ["install"], catch_exceptions=False)
    assert res.exit_code == 0, res.stdout



import datetime

import pytest
import time_machine

from canaille.app import models


def test_model_comparison(testclient, backend):
    foo1 = models.User(
        user_name="foo",
        family_name="foo",
        formatted_name="foo",
    )
    backend.save(foo1)
    backend.reload(foo1)
    bar = models.User(
        user_name="bar",
        family_name="bar",
        formatted_name="bar",
    )
    backend.save(bar)
    backend.reload(bar)
    foo2 = backend.get(models.User, id=foo1.id)

    assert foo1 == foo2
    assert foo1 != bar

    backend.delete(foo1)
    backend.delete(bar)


def test_model_lifecycle(testclient, backend):
    user = models.User(
        user_name="user_name",
        family_name="family_name",
        formatted_name="formatted_name",
    )

    assert not user.id
    assert not backend.query(models.User)
    assert not backend.query(models.User, id=user.id)
    assert not backend.query(models.User, id="invalid")
    assert not backend.get(models.User, id=user.id)

    backend.save(user)
    backend.reload(user)

    assert backend.query(models.User) == [user]
    assert backend.query(models.User, id=user.id) == [user]
    assert not backend.query(models.User, id="invalid")
    assert backend.get(models.User, id=user.id) == user

    user.family_name = "new_family_name"

    assert user.family_name == "new_family_name"

    backend.reload(user)

    assert user.family_name == "family_name"

    backend.delete(user)

    assert not backend.query(models.User, id=user.id)
    assert not backend.get(models.User, id=user.id)

    backend.delete(user)


def test_model_attribute_edition(testclient, backend):
    user = models.User(
        user_name="user_name",
        family_name="family_name",
        formatted_name="formatted_name",
        display_name="display_name",
        emails=["email1@user.test", "email2@user.test"],
    )
    backend.save(user)

    assert user.user_name == "user_name"
    assert user.family_name == "family_name"
    assert user.emails == ["email1@user.test", "email2@user.test"]

    user = backend.get(models.User, id=user.id)
    assert user.user_name == "user_name"
    assert user.family_name == "family_name"
    assert user.emails == ["email1@user.test", "email2@user.test"]

    user.family_name = "new_family_name"
    user.emails = ["email1@user.test"]
    backend.save(user)

    assert user.family_name == "new_family_name"
    assert user.emails == ["email1@user.test"]

    user = backend.get(models.User, id=user.id)
    assert user.family_name == "new_family_name"
    assert user.emails == ["email1@user.test"]

    user.display_name = ""
    assert not user.display_name

    backend.save(user)
    assert not user.display_name

    backend.delete(user)


def test_model_indexation(testclient, backend):
    user = models.User(
        user_name="user_name",
        family_name="family_name",
        formatted_name="formatted_name",
        emails=["email1@user.test", "email2@user.test"],
    )
    backend.save(user)
    backend.reload(user)

    assert backend.get(models.User, family_name="family_name") == user
    assert not backend.get(models.User, family_name="new_family_name")
    assert backend.get(models.User, emails=["email1@user.test"]) == user
    assert backend.get(models.User, emails=["email2@user.test"]) == user
    assert not backend.get(models.User, emails=["email3@user.test"])

    user.family_name = "new_family_name"
    user.emails = ["email2@user.test"]

    assert backend.get(models.User, family_name="family_name") != user
    assert backend.get(models.User, emails=["email1@user.test"]) != user
    assert not backend.get(models.User, emails=["email3@user.test"])

    backend.save(user)
    backend.reload(user)

    assert not backend.get(models.User, family_name="family_name")
    assert backend.get(models.User, family_name="new_family_name") == user
    assert not backend.get(models.User, emails=["email1@user.test"])
    assert backend.get(models.User, emails=["email2@user.test"]) == user
    assert not backend.get(models.User, emails=["email3@user.test"])

    backend.delete(user)

    assert not backend.get(models.User, family_name="family_name")
    assert not backend.get(models.User, family_name="new_family_name")
    assert not backend.get(models.User, emails=["email1@user.test"])
    assert not backend.get(models.User, emails=["email2@user.test"])
    assert not backend.get(models.User, emails=["email3@user.test"])


def test_fuzzy_unique_attribute(user, moderator, admin, backend):
    backend.reload(user)
    backend.reload(moderator)
    backend.reload(admin)

    assert set(backend.query(models.User)) == {user, moderator, admin}
    assert set(backend.fuzzy(models.User, "Jack")) == {moderator}
    assert set(backend.fuzzy(models.User, "Jack", ["formatted_name"])) == {moderator}
    assert set(backend.fuzzy(models.User, "Jack", ["user_name"])) == set()
    assert set(backend.fuzzy(models.User, "Jack", ["user_name", "formatted_name"])) == {
        moderator
    }
    assert set(backend.fuzzy(models.User, "moderator")) == {moderator}
    assert set(backend.fuzzy(models.User, "oderat")) == {moderator}
    assert set(backend.fuzzy(models.User, "oDeRat")) == {moderator}
    assert set(backend.fuzzy(models.User, "ack")) == {moderator}


def test_fuzzy_multiple_attribute(user, moderator, admin, backend):
    backend.reload(user)
    backend.reload(moderator)
    backend.reload(admin)

    assert set(backend.query(models.User)) == {user, moderator, admin}
    assert set(backend.fuzzy(models.User, "jack@doe.test")) == {moderator}
    assert set(backend.fuzzy(models.User, "jack@doe.test", ["emails"])) == {moderator}
    assert set(backend.fuzzy(models.User, "jack@doe.test", ["formatted_name"])) == set()
    assert set(
        backend.fuzzy(models.User, "jack@doe.test", ["emails", "formatted_name"])
    ) == {moderator}
    assert set(backend.fuzzy(models.User, "ack@doe.te")) == {moderator}
    assert set(backend.fuzzy(models.User, "doe.test")) == {user, moderator, admin}


def test_model_references(testclient, user, foo_group, admin, bar_group, backend):
    assert foo_group.members == [user]
    assert user.groups == [foo_group]
    assert foo_group in backend.query(models.Group, members=user)
    assert user in backend.query(models.User, groups=foo_group)

    assert user not in bar_group.members
    assert bar_group not in user.groups
    user.groups = user.groups + [bar_group]
    backend.save(user)
    backend.reload(bar_group)

    assert user in bar_group.members
    assert bar_group in user.groups

    bar_group.members = [admin]
    backend.save(bar_group)
    backend.reload(user)

    assert user not in bar_group.members
    assert bar_group not in user.groups


def test_model_creation_edition_datetime(testclient, backend):
    if "ldap" in backend.__class__.__module__:
        pytest.skip()

    with time_machine.travel("2020-01-01 02:00:00+00:00", tick=False):
        user = models.User(
            user_name="foo",
            family_name="foo",
            formatted_name="foo",
        )
        backend.save(user)
        assert user.created == datetime.datetime(
            2020, 1, 1, 2, tzinfo=datetime.timezone.utc
        )
        assert user.last_modified == datetime.datetime(
            2020, 1, 1, 2, tzinfo=datetime.timezone.utc
        )

    with time_machine.travel("2021-01-01 02:00:00+00:00", tick=False):
        user.family_name = "bar"
        backend.save(user)
        assert user.created == datetime.datetime(
            2020, 1, 1, 2, tzinfo=datetime.timezone.utc
        )
        assert user.last_modified == datetime.datetime(
            2021, 1, 1, 2, tzinfo=datetime.timezone.utc
        )

    backend.delete(user)



from flask import Blueprint
from flask import abort
from flask import current_app
from flask import flash
from flask import g
from flask import redirect
from flask import request
from flask import session
from flask import url_for

from canaille.app import build_hash
from canaille.app.flask import smtp_needed
from canaille.app.i18n import gettext as _
from canaille.app.session import login_user
from canaille.app.session import logout_user
from canaille.app.templating import render_template
from canaille.backends import Backend
from canaille.core.auth import auth_step
from canaille.core.auth import get_user_from_login
from canaille.core.auth import redirect_to_next_auth_step

from ...mails import send_password_initialization_mail
from ...mails import send_password_reset_mail
from ..forms import FirstLoginForm
from ..forms import ForgottenPasswordCodeForm
from ..forms import ForgottenPasswordForm
from ..forms import PasswordForm
from ..forms import PasswordResetForm

bp = Blueprint("password", __name__)


@bp.context_processor
def global_processor():
    return {
        "menu": False,
    }


@bp.route("/auth/password", methods=("GET", "POST"))
@auth_step("password")
def password():
    form = PasswordForm(request.form or None)
    form.render_field_macro_file = "core/partial/login_field.html"

    if not request.form or form.form_control():
        return render_template("core/auth/password.html", form=form)

    if g.auth.user and not g.auth.user.has_password() and current_app.features.has_smtp:
        return redirect(url_for("core.auth.password.firstlogin", user=g.auth.user))

    if not form.validate() or not g.auth.user:
        logout_user()
        flash(_("Login failed, please check your information"), "error")
        return render_template("core/auth/password.html", form=form)

    success, message = Backend.instance.check_user_password(
        g.auth.user, form.password.data
    )
    if not success:
        logout_user()
        current_app.logger.security(
            f"Failed password authentication for {g.auth.user_name}"
        )
        flash(message or _("Login failed, please check your information"), "error")
        return render_template("core/auth/password.html", form=form)

    current_app.logger.security(
        f"Successful password authentication for {g.auth.user_name}"
    )
    g.auth.set_step_finished("password")
    return redirect_to_next_auth_step()


@bp.route("/firstlogin/<user:user>", methods=("GET", "POST"))
def firstlogin(user):
    if user.has_password():
        abort(404)

    form = FirstLoginForm(request.form or None)
    if not request.form:
        return render_template("core/auth/firstlogin.html", form=form)

    form.validate()

    statuses = [
        send_password_initialization_mail(user, email) for email in (user.emails or [])
    ]
    success = all(statuses)
    if success:
        flash(
            _(
                "A password initialization link has been sent at your email address. "
                "You should receive it within a few minutes."
            ),
            "success",
        )
    else:
        flash(_("Could not send the password initialization email"), "error")

    return render_template("core/auth/firstlogin.html", form=form)


@bp.route("/reset", methods=["GET", "POST"])
@smtp_needed()
def forgotten():
    if not current_app.config["CANAILLE"]["ENABLE_PASSWORD_RECOVERY"]:
        abort(404)

    form = ForgottenPasswordForm(request.form)
    if not request.form:
        return render_template("core/auth/forgotten-password.html", form=form)

    item_name = "link" if current_app.features.has_trusted_hosts else "code"

    if not form.validate():
        flash(_(f"Could not send the password reset {item_name}."), "error")
        return render_template("core/auth/forgotten-password.html", form=form)

    user = get_user_from_login(form.login.data)
    success_message = _(
        f"A password reset {item_name} has been sent at your email address. "
        "You should receive it within a few minutes."
    )
    if current_app.config["CANAILLE"]["HIDE_INVALID_LOGINS"] and (
        not user or not user.can_edit_self or user.locked
    ):
        flash(success_message, "success")
        return render_template("core/auth/forgotten-password.html", form=form)

    if not user.can_edit_self:
        flash(
            _(
                "The user '%(user)s' does not have permissions to update their password. "
                "We cannot send a password reset email.",
                user=user.formatted_name,
            ),
            "error",
        )
        return render_template("core/auth/forgotten-password.html", form=form)

    success = True
    for email in user.emails or []:
        if not send_password_reset_mail(user, email):
            success = False
        current_app.logger.security(
            f"Sending a reset password mail to {email} for {user.user_name}"
        )

    if success:
        flash(success_message, "success")
    else:
        flash(
            _("We encountered an issue while we sent the password recovery email."),
            "error",
        )

    if current_app.features.has_trusted_hosts:
        return render_template("core/auth/forgotten-password.html", form=form)
    else:
        return redirect(url_for(".forgotten_code", user=user))


@bp.route("/reset-code/<user:user>", methods=["GET", "POST"])
@smtp_needed()
def forgotten_code(user):
    if (
        not current_app.config["CANAILLE"]["ENABLE_PASSWORD_RECOVERY"]
        or current_app.features.has_trusted_hosts
    ):
        abort(404)

    if not user.can_edit_self:
        flash(
            _(
                "The user '%(user)s' does not have permissions to update their password. ",
                user=user.formatted_name,
            ),
            "error",
        )
        return redirect(url_for(".forgotten"))

    form = ForgottenPasswordCodeForm(request.form)
    if not request.form:
        return render_template("core/auth/forgotten-password-code.html", form=form)

    if not form.validate() or not user.is_email_or_sms_otp_valid(form.code.data):
        flash(_("Invalid code."), "error")
        return render_template("core/auth/forgotten-password-code.html", form=form)

    return redirect(url_for(".reset", user=user, token=form.code.data))


@bp.route("/reset/<user:user>/<token>", methods=["GET", "POST"])
def reset(user, token):
    if not current_app.config["CANAILLE"]["ENABLE_PASSWORD_RECOVERY"]:
        abort(404)
    form = PasswordResetForm(request.form)

    if current_app.features.has_trusted_hosts:
        token = build_hash(token)

    if not user or not user.is_email_or_sms_otp_valid(token):
        item_name = "link" if current_app.features.has_trusted_hosts else "code"
        flash(
            _(f"The password reset {item_name} that brought you here was invalid."),
            "error",
        )
        return redirect(url_for("core.account.index"))

    if not request.form or not form.validate():
        return render_template(
            "core/auth/reset-password.html", form=form, user=user, token=token
        )

    Backend.instance.set_user_password(user, form.password.data)
    user.clear_otp()
    Backend.instance.save(user)
    login_user(user)

    flash(_("Your password has been updated successfully"), "success")
    return redirect(
        session.pop(
            "redirect-after-login",
            url_for("core.account.profile_edition", edited_user=user),
        )
    )



"""Add login metadata to AuthorizationCode.

Revision ID: 1743675164
Revises: 1742298790
Create Date: 2025-04-03 12:12:44.237218

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

import canaille.backends.sql.utils

# revision identifiers, used by Alembic.
revision: str = "1743675164"
down_revision: str | None = "1742298790"
branch_labels: str | Sequence[str] | None = ()
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "authorization_code",
        sa.Column(
            "auth_time",
            canaille.backends.sql.utils.TZDateTime(timezone=True),
            nullable=True,
        ),
    )
    op.add_column("authorization_code", sa.Column("acr", sa.String(), nullable=True))
    op.add_column("authorization_code", sa.Column("amr", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("authorization_code", "amr")
    op.drop_column("authorization_code", "acr")
    op.drop_column("authorization_code", "auth_time")
    # ### end Alembic commands ###



import datetime
from unittest import mock

import time_machine
from flask import url_for

from canaille.core.endpoints.account import EmailConfirmationPayload
from canaille.core.endpoints.account import RegistrationPayload


def test_confirmation_disabled_email_editable(testclient, backend, logged_user):
    """If email confirmation is disabled, users should be able to pick any email."""
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = False

    res = testclient.get("/profile/user")
    assert "readonly" not in res.form["emails-0"].attrs
    assert not any(field.id == "add_email" for field in res.form.fields["action"])

    res = res.form.submit(name="fieldlist_add", value="emails-0")
    res.form["emails-0"] = "email1@mydomain.test"
    res.form["emails-1"] = "email2@mydomain.test"

    res = res.form.submit(name="action", value="edit-profile")
    assert res.flashes == [("success", "Profile updated successfully.")]
    res = res.follow()

    backend.reload(logged_user)

    assert logged_user.emails == ["email1@mydomain.test", "email2@mydomain.test"]


def test_confirmation_unset_smtp_disabled_email_editable(
    testclient, backend, logged_admin, user
):
    """If email confirmation is unset and no SMTP server has been configured, then email confirmation cannot be enabled, thus users must be able to pick any email."""
    testclient.app.config["CANAILLE"]["SMTP"] = None
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = None

    res = testclient.get("/profile/user")
    assert "readonly" not in res.form["emails-0"].attrs
    assert not any(field.id == "add_email" for field in res.form.fields["action"])

    res = res.form.submit(name="fieldlist_add", value="emails-0")
    res.form["emails-0"] = "email1@mydomain.test"
    res.form["emails-1"] = "email2@mydomain.test"

    res = res.form.submit(name="action", value="edit-profile")
    assert res.flashes == [("success", "Profile updated successfully.")]
    res = res.follow()

    backend.reload(user)
    assert user.emails == ["email1@mydomain.test", "email2@mydomain.test"]


def test_confirmation_enabled_smtp_disabled_readonly(testclient, backend, logged_user):
    """If email confirmation is enabled and no SMTP server is configured, this might be a misconfiguration, or a temporary SMTP disabling.

    In doubt, users cannot edit their emails.
    """
    testclient.app.config["CANAILLE"]["SMTP"] = None
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = True

    res = testclient.get("/profile/user")
    assert "readonly" in res.forms["emailconfirmationform"]["old_emails-0"].attrs
    assert "emails-0" not in res.forms["baseform"].fields

    res.forms["emailconfirmationform"]["old_emails-0"] = "email1@mydomain.test"
    assert "action" not in res.forms["emailconfirmationform"].fields


def test_confirmation_unset_smtp_enabled_email_admin_editable(
    testclient, backend, logged_admin, user
):
    """Administrators should be able to edit user email addresses, even when email confirmation is unset and SMTP is configured."""
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = None

    res = testclient.get("/profile/user")
    assert "readonly" not in res.form["emails-0"].attrs
    assert not any(field.id == "add_email" for field in res.form.fields["action"])

    res = res.form.submit(name="fieldlist_add", value="emails-0")
    res.form["emails-0"] = "email1@mydomain.test"
    res.form["emails-1"] = "email2@mydomain.test"

    res = res.form.submit(name="action", value="edit-profile")
    assert res.flashes == [("success", "Profile updated successfully.")]
    res = res.follow()

    backend.reload(user)
    assert user.emails == ["email1@mydomain.test", "email2@mydomain.test"]


def test_confirmation_enabled_smtp_disabled_admin_editable(
    testclient, backend, logged_admin, user
):
    """Administrators should be able to edit user email addresses, even when email confirmation is enabled and SMTP is disabled."""
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = True
    testclient.app.config["CANAILLE"]["SMTP"] = None

    res = testclient.get("/profile/user")
    assert "readonly" not in res.form["emails-0"].attrs
    assert not any(field.id == "add_email" for field in res.form.fields["action"])

    res = res.form.submit(name="fieldlist_add", value="emails-0")
    res.form["emails-0"] = "email1@mydomain.test"
    res.form["emails-1"] = "email2@mydomain.test"

    res = res.form.submit(name="action", value="edit-profile")
    assert res.flashes == [("success", "Profile updated successfully.")]
    res = res.follow()

    backend.reload(user)
    assert user.emails == ["email1@mydomain.test", "email2@mydomain.test"]


def test_confirmation_unset_smtp_enabled_email_user_validation(
    smtpd, testclient, backend, user
):
    """If email confirmation is unset and there is a SMTP server configured, then users emails should be validated by sending a confirmation email."""
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = None

    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get("/login")
        res.form["login"] = "user"
        res = res.form.submit().follow()
        res.form["password"] = "correct horse battery staple"
        res = res.form.submit()

    with time_machine.travel("2020-01-01 02:00:00+00:00", tick=False):
        res = testclient.get("/profile/user")

    assert "readonly" in res.forms["emailconfirmationform"]["old_emails-0"].attrs

    with time_machine.travel("2020-01-01 02:00:00+00:00", tick=False):
        res.forms["emailconfirmationform"]["new_email"] = "new_email@mydomain.test"
        res = res.forms["emailconfirmationform"].submit(
            name="action", value="add_email"
        )

    assert res.flashes == [
        (
            "success",
            "An email has been sent to the email address. "
            "Please check your inbox and click on the verification link it contains",
        )
    ]

    email_confirmation = EmailConfirmationPayload(
        "2020-01-01T02:00:00+00:00",
        "user",
        "new_email@mydomain.test",
    )
    email_confirmation_url = url_for(
        "core.account.email_confirmation",
        data=email_confirmation.b64(),
        hash=email_confirmation.build_hash(),
        _external=True,
    )

    assert len(smtpd.messages) == 1
    email_content = (
        str(smtpd.messages[0].get_payload()[0]).replace("=\n", "").replace("=3D", "=")
    )
    assert email_confirmation_url in email_content

    with time_machine.travel("2020-01-01 03:00:00+00:00", tick=False):
        res = testclient.get(email_confirmation_url)

    assert ("success", "Your email address have been confirmed.") in res.flashes
    backend.reload(user)
    assert "new_email@mydomain.test" in user.emails


def test_confirmation_unset_smtp_enabled_email_user_validation_htmx(
    smtpd, testclient, backend, user
):
    """Test that the edition of the form with HTMX don't cause errors."""
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = None

    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get("/login")
        res.form["login"] = "user"
        res = res.form.submit().follow()
        res.form["password"] = "correct horse battery staple"
        res = res.form.submit()

    with time_machine.travel("2020-01-01 02:00:00+00:00", tick=False):
        res = testclient.get("/profile/user")

    assert "readonly" in res.forms["emailconfirmationform"]["old_emails-0"].attrs

    with time_machine.travel("2020-01-01 02:00:00+00:00", tick=False):
        res.forms["emailconfirmationform"]["new_email"] = "new_email@mydomain.test"
        res = res.forms["emailconfirmationform"].submit(
            name="action",
            value="add_email",
            headers={
                "HX-Request": "true",
                "HX-Trigger-Name": "new_email",
            },
        )
        assert "</html>" not in res.text


def test_confirmation_invalid_link(testclient, backend, user):
    """Random confirmation links should fail."""
    res = testclient.get("/email-confirmation/invalid/invalid")
    assert (
        "error",
        "The email confirmation link that brought you here is invalid.",
    ) in res.flashes


def test_confirmation_mail_form_failed(testclient, backend, user):
    """Tests when an error happens during the mail sending."""
    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get("/login")
        res.form["login"] = "user"
        res = res.form.submit().follow()
        res.form["password"] = "correct horse battery staple"
        res = res.form.submit()

    with time_machine.travel("2020-01-01 02:00:00+00:00", tick=False):
        res = testclient.get("/profile/user")

    assert "readonly" in res.forms["emailconfirmationform"]["old_emails-0"].attrs

    with time_machine.travel("2020-01-01 02:00:00+00:00", tick=False):
        res.forms["emailconfirmationform"]["new_email"] = "invalid"
        res = res.forms["emailconfirmationform"].submit(
            name="action", value="add_email"
        )

    assert res.flashes == [("error", "Email addition failed.")]
    backend.reload(user)
    assert user.emails == ["john@doe.test"]


@mock.patch("smtplib.SMTP")
def test_confirmation_mail_send_failed(SMTP, smtpd, testclient, backend, user):
    """Tests when an error happens during the mail sending."""
    SMTP.side_effect = mock.Mock(side_effect=OSError("unit test mail error"))
    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get("/login")
        res.form["login"] = "user"
        res = res.form.submit().follow()
        res.form["password"] = "correct horse battery staple"
        res = res.form.submit()

    with time_machine.travel("2020-01-01 02:00:00+00:00", tick=False):
        res = testclient.get("/profile/user")

    assert "readonly" in res.forms["emailconfirmationform"]["old_emails-0"].attrs

    with time_machine.travel("2020-01-01 02:00:00+00:00", tick=False):
        res.forms["emailconfirmationform"]["new_email"] = "new_email@mydomain.test"
        res = res.forms["emailconfirmationform"].submit(
            name="action", value="add_email", expect_errors=True
        )

    assert res.flashes == [("error", "Could not send the verification email")]
    backend.reload(user)
    assert user.emails == ["john@doe.test"]


def test_confirmation_expired_link(testclient, backend, user):
    """Expired valid confirmation links should fail."""
    email_confirmation = EmailConfirmationPayload(
        "2020-01-01T01:00:00+00:00",
        "user",
        "new_email@mydomain.test",
    )
    email_confirmation_url = url_for(
        "core.account.email_confirmation",
        data=email_confirmation.b64(),
        hash=email_confirmation.build_hash(),
        _external=True,
    )

    with time_machine.travel("2021-01-01 01:00:00+00:00", tick=False):
        res = testclient.get(email_confirmation_url)

    assert (
        "error",
        "The email confirmation link that brought you here has expired.",
    ) in res.flashes
    backend.reload(user)
    assert "new_email@mydomain.test" not in user.emails


def test_confirmation_invalid_hash_link(testclient, backend, user):
    """Confirmation link with invalid hashes should fail."""
    email_confirmation = EmailConfirmationPayload(
        "2020-01-01T01:00:00+00:00",
        "user",
        "new_email@mydomain.test",
    )
    email_confirmation_url = url_for(
        "core.account.email_confirmation",
        data=email_confirmation.b64(),
        hash="invalid",
        _external=True,
    )

    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get(email_confirmation_url)

    assert (
        "error",
        "The invitation link that brought you here was invalid.",
    ) in res.flashes
    backend.reload(user)
    assert "new_email@mydomain.test" not in user.emails


def test_confirmation_invalid_user_link(testclient, backend, user):
    """Confirmation link about an unexisting user should fail.

    For instance, when the user account has been deleted between the
    mail is sent and the link is clicked.
    """
    email_confirmation = EmailConfirmationPayload(
        "2020-01-01T01:00:00+00:00",
        "invalid-user",
        "new_email@mydomain.test",
    )
    email_confirmation_url = url_for(
        "core.account.email_confirmation",
        data=email_confirmation.b64(),
        hash=email_confirmation.build_hash(),
        _external=True,
    )

    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get(email_confirmation_url)

    assert (
        "error",
        "The email confirmation link that brought you here is invalid.",
    ) in res.flashes
    backend.reload(user)
    assert "new_email@mydomain.test" not in user.emails


def test_confirmation_email_already_confirmed_link(testclient, backend, user, admin):
    """Clicking twice on a confirmation link should fail."""
    email_confirmation = EmailConfirmationPayload(
        "2020-01-01T01:00:00+00:00",
        "user",
        "john@doe.test",
    )
    email_confirmation_url = url_for(
        "core.account.email_confirmation",
        data=email_confirmation.b64(),
        hash=email_confirmation.build_hash(),
        _external=True,
    )

    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get(email_confirmation_url)

    assert (
        "error",
        "This address email have already been confirmed.",
    ) in res.flashes
    backend.reload(user)
    assert "new_email@mydomain.test" not in user.emails


def test_confirmation_email_already_used_link(testclient, backend, user, admin):
    """Confirmation link should fail if the target email is already associated to another account.

    For instance, if an administrator already put this email to someone
    else's profile.
    """
    email_confirmation = EmailConfirmationPayload(
        "2020-01-01T01:00:00+00:00",
        "user",
        "jane@doe.test",
    )
    email_confirmation_url = url_for(
        "core.account.email_confirmation",
        data=email_confirmation.b64(),
        hash=email_confirmation.build_hash(),
        _external=True,
    )

    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get(email_confirmation_url)

    assert (
        "error",
        "This address email is already associated with another account.",
    ) in res.flashes
    backend.reload(user)
    assert "new_email@mydomain.test" not in user.emails


def test_delete_email(testclient, logged_user, backend):
    """Tests that user can deletes its emails unless they have only one left."""
    res = testclient.get("/profile/user")
    assert "email_remove" not in res.forms["emailconfirmationform"].fields

    logged_user.emails = logged_user.emails + ["new@email.test"]
    backend.save(logged_user)
    res = testclient.get("/profile/user")
    assert "email_remove" in res.forms["emailconfirmationform"].fields

    res = res.forms["emailconfirmationform"].submit(
        name="email_remove", value="new@email.test"
    )
    assert res.flashes == [("success", "The email have been successfully deleted.")]

    backend.reload(logged_user)
    assert logged_user.emails == ["john@doe.test"]


def test_delete_wrong_email(testclient, logged_user, backend):
    """Tests that removing an already removed email do not produce anything."""
    logged_user.emails = logged_user.emails + ["new@email.test"]
    backend.save(logged_user)

    res = testclient.get("/profile/user")

    res1 = res.forms["emailconfirmationform"].submit(
        name="email_remove", value="new@email.test"
    )
    assert res1.flashes == [("success", "The email have been successfully deleted.")]

    res2 = res.forms["emailconfirmationform"].submit(
        name="email_remove", value="new@email.test"
    )
    assert res2.flashes == [("error", "Email deletion failed.")]

    backend.reload(logged_user)
    assert logged_user.emails == ["john@doe.test"]


def test_delete_last_email(testclient, logged_user, backend):
    """Tests that users cannot remove their last email address."""
    logged_user.emails = logged_user.emails + ["new@email.test"]
    backend.save(logged_user)

    res = testclient.get("/profile/user")

    res1 = res.forms["emailconfirmationform"].submit(
        name="email_remove", value="new@email.test"
    )
    assert res1.flashes == [("success", "The email have been successfully deleted.")]

    res2 = res.forms["emailconfirmationform"].submit(
        name="email_remove", value="john@doe.test"
    )
    assert res2.flashes == [("error", "Email deletion failed.")]

    backend.reload(logged_user)
    assert logged_user.emails == ["john@doe.test"]


def test_edition_forced_mail(testclient, logged_user, backend):
    """Tests that users that must perform email verification cannot force the profile form."""
    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    testclient.post(
        "/profile/user",
        {
            "csrf_token": form["csrf_token"].value,
            "emails-0": "new@email.test",
            "action": "edit-profile",
        },
    )

    backend.reload(logged_user)
    assert logged_user.emails == ["john@doe.test"]


def test_invitation_form_mail_field_readonly(testclient):
    """Tests that the email field is readonly in the invitation form creation if email confirmation is enabled."""
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = True

    payload = RegistrationPayload(
        datetime.datetime.now(datetime.timezone.utc).isoformat(),
        "someoneelse",
        False,
        "someone@mydomain.test",
        [],
    )
    hash = payload.build_hash()
    b64 = payload.b64()

    res = testclient.get(f"/register/{b64}/{hash}")
    assert "readonly" in res.form["emails-0"].attrs


def test_invitation_form_mail_field_writable(testclient):
    """Tests that the email field is writable in the invitation form creation if email confirmation is disabled."""
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = False

    payload = RegistrationPayload(
        datetime.datetime.now(datetime.timezone.utc).isoformat(),
        "someoneelse",
        False,
        "someone@mydomain.test",
        [],
    )
    hash = payload.build_hash()
    b64 = payload.b64()

    res = testclient.get(f"/register/{b64}/{hash}")
    assert "readonly" not in res.form["emails-0"].attrs
